"""
***************************************************************************
    menus.py
    ---------------------
    Date                 : February 2016
    Copyright            : (C) 2016 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "February 2016"
__copyright__ = "(C) 2016, Victor Olaya"

import os

from qgis.core import (
    QgsApplication,
    QgsMessageLog,
    QgsProcessingAlgorithm,
    QgsStringUtils,
)
from qgis.gui import QgsGui
from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QApplication, QMenu, QToolButton
from qgis.utils import iface

from processing.core.Processing import Processing
from processing.core.ProcessingConfig import ProcessingConfig, Setting
from processing.gui.algorithm_widget import AlgorithmWidget
from processing.gui.AlgorithmExecutor import execute
from processing.gui.MessageBarProgress import MessageBarProgress
from processing.gui.MessageDialog import MessageDialog
from processing.gui.Postprocessing import handleAlgorithmResults
from processing.tools import dataobjects

algorithmsToolbar = None
menusSettingsGroup = "Menus"
defaultMenuEntries = {}
toolBarButtons = []
toolButton = None
toolButtonAction = None


def initMenusAndToolbars():
    global defaultMenuEntries, toolBarButtons, toolButton, toolButtonAction
    vectorMenu = iface.vectorMenu().title()
    analysisToolsMenu = (
        vectorMenu
        + "/"
        + QCoreApplication.translate("ProcessingPlugin", "&Analysis Tools")
    )
    defaultMenuEntries.update(
        {
            "native:distancematrix": analysisToolsMenu,
            "native:sumlinelengths": analysisToolsMenu,
            "native:countpointsinpolygon": analysisToolsMenu,
            "native:listuniquevalues": analysisToolsMenu,
            "native:basicstatisticsforfields": analysisToolsMenu,
            "native:nearestneighbouranalysis": analysisToolsMenu,
            "native:meancoordinates": analysisToolsMenu,
            "native:lineintersections": analysisToolsMenu,
        }
    )
    researchToolsMenu = (
        vectorMenu
        + "/"
        + QCoreApplication.translate("ProcessingPlugin", "&Research Tools")
    )
    defaultMenuEntries.update(
        {
            "native:creategrid": researchToolsMenu,
            "native:randomselection": researchToolsMenu,
            "native:randomselectionwithinsubsets": researchToolsMenu,
            "native:randompointsinextent": researchToolsMenu,
            "qgis:randompointsinlayerbounds": researchToolsMenu,
            "native:randompointsinpolygons": researchToolsMenu,
            "qgis:randompointsinsidepolygons": researchToolsMenu,
            "native:randompointsonlines": researchToolsMenu,
            "qgis:regularpoints": researchToolsMenu,
            "native:selectbylocation": researchToolsMenu,
            "native:selectwithindistance": researchToolsMenu,
            "native:polygonfromlayerextent": researchToolsMenu,
        }
    )
    geoprocessingToolsMenu = (
        vectorMenu
        + "/"
        + QCoreApplication.translate("ProcessingPlugin", "&Geoprocessing Tools")
    )
    defaultMenuEntries.update(
        {
            "native:buffer": geoprocessingToolsMenu,
            "native:convexhull": geoprocessingToolsMenu,
            "native:intersection": geoprocessingToolsMenu,
            "native:union": geoprocessingToolsMenu,
            "native:symmetricaldifference": geoprocessingToolsMenu,
            "native:clip": geoprocessingToolsMenu,
            "native:difference": geoprocessingToolsMenu,
            "native:dissolve": geoprocessingToolsMenu,
            "qgis:eliminateselectedpolygons": geoprocessingToolsMenu,
        }
    )
    geometryToolsMenu = (
        vectorMenu
        + "/"
        + QCoreApplication.translate("ProcessingPlugin", "G&eometry Tools")
    )
    defaultMenuEntries.update(
        {
            "native:checkvalidity": geometryToolsMenu,
            "native:exportaddgeometrycolumns": geometryToolsMenu,
            "native:centroids": geometryToolsMenu,
            "native:delaunaytriangulation": geometryToolsMenu,
            "native:voronoipolygons": geometryToolsMenu,
            "native:simplifygeometries": geometryToolsMenu,
            "native:densifygeometries": geometryToolsMenu,
            "native:multiparttosingleparts": geometryToolsMenu,
            "native:collect": geometryToolsMenu,
            "native:polygonstolines": geometryToolsMenu,
            "qgis:linestopolygons": geometryToolsMenu,
            "native:extractvertices": geometryToolsMenu,
        }
    )
    managementToolsMenu = (
        vectorMenu
        + "/"
        + QCoreApplication.translate("ProcessingPlugin", "&Data Management Tools")
    )
    defaultMenuEntries.update(
        {
            "native:reprojectlayer": managementToolsMenu,
            "native:joinattributesbylocation": managementToolsMenu,
            "native:splitvectorlayer": managementToolsMenu,
            "native:mergevectorlayers": managementToolsMenu,
            "native:createspatialindex": managementToolsMenu,
        }
    )

    rasterMenu = iface.rasterMenu().title()
    defaultMenuEntries.update({"native:alignrasters": rasterMenu})
    projectionsMenu = (
        rasterMenu + "/" + QCoreApplication.translate("ProcessingPlugin", "Projections")
    )
    defaultMenuEntries.update(
        {
            "gdal:warpreproject": projectionsMenu,
            "gdal:extractprojection": projectionsMenu,
            "gdal:assignprojection": projectionsMenu,
        }
    )
    conversionMenu = (
        rasterMenu + "/" + QCoreApplication.translate("ProcessingPlugin", "Conversion")
    )
    defaultMenuEntries.update(
        {
            "gdal:rasterize": conversionMenu,
            "gdal:polygonize": conversionMenu,
            "gdal:translate": conversionMenu,
            "gdal:rgbtopct": conversionMenu,
            "gdal:pcttorgb": conversionMenu,
        }
    )
    extractionMenu = (
        rasterMenu + "/" + QCoreApplication.translate("ProcessingPlugin", "Extraction")
    )
    defaultMenuEntries.update(
        {
            "gdal:contour": extractionMenu,
            "gdal:cliprasterbyextent": extractionMenu,
            "gdal:cliprasterbymasklayer": extractionMenu,
        }
    )
    analysisMenu = (
        rasterMenu + "/" + QCoreApplication.translate("ProcessingPlugin", "Analysis")
    )
    defaultMenuEntries.update(
        {
            "gdal:sieve": analysisMenu,
            "gdal:nearblack": analysisMenu,
            "gdal:fillnodata": analysisMenu,
            "gdal:proximity": analysisMenu,
            "gdal:griddatametrics": analysisMenu,
            "gdal:gridaverage": analysisMenu,
            "gdal:gridinversedistance": analysisMenu,
            "gdal:gridnearestneighbor": analysisMenu,
            "gdal:aspect": analysisMenu,
            "gdal:hillshade": analysisMenu,
            "gdal:roughness": analysisMenu,
            "gdal:slope": analysisMenu,
            "gdal:tpitopographicpositionindex": analysisMenu,
            "gdal:triterrainruggednessindex": analysisMenu,
        }
    )
    miscMenu = (
        rasterMenu
        + "/"
        + QCoreApplication.translate("ProcessingPlugin", "Miscellaneous")
    )
    defaultMenuEntries.update(
        {
            "gdal:buildvirtualraster": miscMenu,
            "gdal:merge": miscMenu,
            "gdal:gdalinfo": miscMenu,
            "gdal:overviews": miscMenu,
            "gdal:tileindex": miscMenu,
        }
    )

    toolBarButtons = ["native:selectbylocation", "native:selectwithindistance"]

    toolbar = iface.selectionToolBar()
    toolButton = QToolButton(toolbar)
    toolButton.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)
    toolButtonAction = toolbar.addWidget(toolButton)
    toolButtonAction.setObjectName("selectByToolButton")


if iface is not None:
    initMenusAndToolbars()


def initializeMenus():
    for m in defaultMenuEntries.keys():
        alg = QgsApplication.processingRegistry().algorithmById(m)
        if alg is None or alg.id() != m:
            QgsMessageLog.logMessage(
                QCoreApplication.translate(
                    "ProcessingPlugin", "Invalid algorithm ID for menu: {}"
                ).format(m),
                QCoreApplication.translate("ProcessingPlugin", "Processing"),
            )

    for provider in QgsApplication.processingRegistry().providers():
        for alg in provider.algorithms():
            d = defaultMenuEntries.get(alg.id(), "")
            setting = Setting(menusSettingsGroup, "MENU_" + alg.id(), "Menu path", d)
            ProcessingConfig.addSetting(setting)
            setting = Setting(
                menusSettingsGroup, "BUTTON_" + alg.id(), "Add button", False
            )
            ProcessingConfig.addSetting(setting)
            setting = Setting(
                menusSettingsGroup,
                "ICON_" + alg.id(),
                "Icon",
                "",
                valuetype=Setting.FILE,
            )
            ProcessingConfig.addSetting(setting)

    ProcessingConfig.readSettings()


def updateMenus():
    removeMenus()
    QCoreApplication.processEvents()
    createMenus()


def createMenus():
    for alg in QgsApplication.processingRegistry().algorithms():
        menuPath = ProcessingConfig.getSetting("MENU_" + alg.id())
        addButton = ProcessingConfig.getSetting("BUTTON_" + alg.id())
        icon = ProcessingConfig.getSetting("ICON_" + alg.id())
        if icon and os.path.exists(icon):
            icon = QIcon(icon)
        else:
            icon = None
        if menuPath:
            paths = menuPath.split("/")
            subMenuName = paths[-1] if len(paths) > 1 else ""
            addAlgorithmEntry(
                alg, paths[0], subMenuName, addButton=addButton, icon=icon
            )


def removeMenus():
    for alg in QgsApplication.processingRegistry().algorithms():
        menuPath = ProcessingConfig.getSetting("MENU_" + alg.id())
        if menuPath:
            paths = menuPath.split("/")
            removeAlgorithmEntry(alg, paths[0], paths[-1])


def addAlgorithmEntry(
    alg, menuName, submenuName, actionText=None, icon=None, addButton=False
):
    if actionText is None:
        if (QgsGui.higFlags() & QgsGui.HigFlag.HigMenuTextIsTitleCase) and not (
            alg.flags() & QgsProcessingAlgorithm.Flag.FlagDisplayNameIsLiteral
        ):
            alg_title = QgsStringUtils.capitalize(
                alg.displayName(), QgsStringUtils.Capitalization.TitleCase
            )
        else:
            alg_title = alg.displayName()
        actionText = alg_title + QCoreApplication.translate("Processing", "…")
    action = QAction(icon or alg.icon(), actionText, iface.mainWindow())
    alg_id = alg.id()
    action.setData(alg_id)
    action.triggered.connect(lambda: _executeAlgorithm(alg_id))
    action.setObjectName("mProcessingUserMenu_%s" % alg_id)

    if menuName:
        menu = getMenu(menuName, iface.mainWindow().menuBar())
        if submenuName:
            submenu = getMenu(submenuName, menu)
            submenu.addAction(action)
        else:
            menu.addAction(action)

    if addButton:
        global algorithmsToolbar
        if algorithmsToolbar is None:
            algorithmsToolbar = iface.addToolBar(
                QCoreApplication.translate("MainWindow", "Processing Algorithms")
            )
            algorithmsToolbar.setObjectName("ProcessingAlgorithms")
            algorithmsToolbar.setToolTip(
                QCoreApplication.translate(
                    "MainWindow", "Processing Algorithms Toolbar"
                )
            )
        algorithmsToolbar.addAction(action)


def removeAlgorithmEntry(alg, menuName, submenuName, delButton=True):
    if menuName:
        menu = getMenu(menuName, iface.mainWindow().menuBar())
        subMenu = getMenu(submenuName, menu)
        action = findAction(subMenu.actions(), alg)
        if action is not None:
            subMenu.removeAction(action)

        if len(subMenu.actions()) == 0:
            subMenu.deleteLater()

    if delButton:
        global algorithmsToolbar
        if algorithmsToolbar is not None:
            action = findAction(algorithmsToolbar.actions(), alg)
            if action is not None:
                algorithmsToolbar.removeAction(action)


def _executeAlgorithm(alg_id):
    alg = QgsApplication.processingRegistry().createAlgorithmById(alg_id)
    if alg is None:
        dlg = MessageDialog()
        dlg.setTitle(
            QCoreApplication.translate("ProcessingPlugin", "Missing Algorithm")
        )
        dlg.setMessage(
            QCoreApplication.translate(
                "ProcessingPlugin",
                'The algorithm "{}" is no longer available. (Perhaps a plugin was uninstalled?)',
            ).format(alg_id)
        )
        dlg.exec()
        return

    ok, message = alg.canExecute()
    if not ok:
        dlg = MessageDialog()
        dlg.setTitle(
            QCoreApplication.translate("ProcessingPlugin", "Missing Dependency")
        )
        dlg.setMessage(
            QCoreApplication.translate(
                "ProcessingPlugin",
                "<h3>Missing dependency. This algorithm cannot be run </h3>\n{0}",
            ).format(message)
        )
        dlg.exec()
        return

    if (alg.countVisibleParameters()) > 0:
        widget = alg.createCustomParametersWidget(parent=iface.mainWindow())
        if not widget:
            widget = AlgorithmWidget(alg)
        widget.exec()
    else:
        feedback = MessageBarProgress()
        context = dataobjects.createContext(feedback)
        parameters = {}
        ret, results = execute(alg, parameters, context, feedback)
        handleAlgorithmResults(alg, context, feedback)
        feedback.close()


def getMenu(name, parent):
    menus = [c for c in parent.children() if isinstance(c, QMenu) and c.title() == name]
    if menus:
        return menus[0]
    else:
        menu = parent.addMenu(name)
        menu.setObjectName(name)
        return menu


def findAction(actions, alg):
    for action in actions:
        if (isinstance(alg, str) and action.data() == alg) or (
            isinstance(alg, QgsProcessingAlgorithm) and action.data() == alg.id()
        ):
            return action
    return None


def addToolBarButton(index, algId, icon=None, tooltip=None):
    alg = QgsApplication.processingRegistry().algorithmById(algId)
    if alg is None or alg.id() != algId:
        assert False, algId

    if tooltip is None:
        if (QgsGui.higFlags() & QgsGui.HigFlag.HigMenuTextIsTitleCase) and not (
            alg.flags() & QgsProcessingAlgorithm.Flag.FlagDisplayNameIsLiteral
        ):
            tooltip = QgsStringUtils.capitalize(
                alg.displayName(), QgsStringUtils.Capitalization.TitleCase
            )
        else:
            tooltip = alg.displayName()

    action = QAction(icon or alg.icon(), tooltip, iface.mainWindow())
    algId = alg.id()
    action.setData(algId)
    action.triggered.connect(lambda: _executeAlgorithm(algId))
    action.setObjectName("mProcessingAlg_%s" % algId)

    toolButton.addAction(action)
    if index == 0:
        toolButton.setDefaultAction(action)


def createButtons():
    toolbar = iface.selectionToolBar()
    for index, algId in enumerate(toolBarButtons):
        addToolBarButton(index, algId)


def removeButtons():
    iface.selectionToolBar().removeAction(toolButtonAction)



# The PEP 484 type hints stub file for the _analysis module.
#
# Generated by SIP 6.15.3


import collections, re, typing, enum

try:
    from warnings import deprecated
except ImportError:
    pass

import PyQt6.sip

from PyQt6 import QtCore
from PyQt6 import QtGui
from PyQt6 import QtXml
from PyQt6 import QtNetwork
from PyQt6 import QtWidgets
from PyQt6 import QtSql
from PyQt6 import QtPrintSupport
from PyQt6 import QtPositioning
from PyQt6 import QtMultimedia
from PyQt6 import QtSerialPort
from qgis import _core

# Support for QDate, QDateTime and QTime.
import datetime

# Convenient type aliases.
PYQT_SIGNAL = typing.Union[QtCore.pyqtSignal, QtCore.pyqtBoundSignal]
PYQT_SLOT = typing.Union[collections.abc.Callable[..., Any], QtCore.pyqtBoundSignal]


class QgsAnalysis(PyQt6.sip.wrapper):

    @staticmethod
    def geometryCheckRegistry() -> 'QgsGeometryCheckRegistry|None': ...
    @staticmethod
    def instance() -> 'QgsAnalysis|None': ...


class QgsGcpGeometryTransformer(_core.QgsAbstractGeometryTransformer):

    @typing.overload
    def __init__(self, gcpTransformer: 'QgsGcpTransformerInterface|None') -> None: ...
    @typing.overload
    def __init__(self, method: 'QgsGcpTransformerInterface.TransformMethod', sourceCoordinates: collections.abc.Iterable[_core.QgsPointXY], destinationCoordinates: collections.abc.Iterable[_core.QgsPointXY]) -> None: ...

    def setGcpTransformer(self, transformer: 'QgsGcpTransformerInterface|None') -> None: ...
    def gcpTransformer(self) -> 'QgsGcpTransformerInterface|None': ...
    def transform(self, geometry: _core.QgsGeometry, feedback: _core.QgsFeedback|None = ...) -> typing.Tuple[_core.QgsGeometry, bool]: ...
    def transformPoint(self, x: float, y: float, z: float, m: float) -> typing.Tuple[bool, float, float, float, float]: ...


class QgsGcpPoint(PyQt6.sip.wrapper):

    class PointType(enum.Enum):
        Source = ... # type: QgsGcpPoint.PointType
        Destination = ... # type: QgsGcpPoint.PointType

    @typing.overload
    def __init__(self, sourcePoint: _core.QgsPointXY, destinationPoint: _core.QgsPointXY, destinationPointCrs: _core.QgsCoordinateReferenceSystem, enabled: bool = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGcpPoint') -> None: ...

    def __ne__(self, other: object): ...
    def __eq__(self, other: object): ...
    def setEnabled(self, enabled: bool) -> None: ...
    def isEnabled(self) -> bool: ...
    def transformedDestinationPoint(self, targetCrs: _core.QgsCoordinateReferenceSystem, context: _core.QgsCoordinateTransformContext) -> _core.QgsPointXY: ...
    def setDestinationPointCrs(self, crs: _core.QgsCoordinateReferenceSystem) -> None: ...
    def destinationPointCrs(self) -> _core.QgsCoordinateReferenceSystem: ...
    def setDestinationPoint(self, point: _core.QgsPointXY) -> None: ...
    def destinationPoint(self) -> _core.QgsPointXY: ...
    def setSourcePoint(self, point: _core.QgsPointXY) -> None: ...
    def sourcePoint(self) -> _core.QgsPointXY: ...


class QgsGcpTransformerInterface(PyQt6.sip.wrapper):

    class TransformMethod(enum.IntEnum):
        Linear = ... # type: QgsGcpTransformerInterface.TransformMethod
        Helmert = ... # type: QgsGcpTransformerInterface.TransformMethod
        PolynomialOrder1 = ... # type: QgsGcpTransformerInterface.TransformMethod
        PolynomialOrder2 = ... # type: QgsGcpTransformerInterface.TransformMethod
        PolynomialOrder3 = ... # type: QgsGcpTransformerInterface.TransformMethod
        ThinPlateSpline = ... # type: QgsGcpTransformerInterface.TransformMethod
        Projective = ... # type: QgsGcpTransformerInterface.TransformMethod
        InvalidTransform = ... # type: QgsGcpTransformerInterface.TransformMethod

    staticMetaObject = ... # type: QtCore.QMetaObject

    def __init__(self) -> None: ...

    @staticmethod
    def createFromParameters(method: 'QgsGcpTransformerInterface.TransformMethod', sourceCoordinates: collections.abc.Iterable[_core.QgsPointXY], destinationCoordinates: collections.abc.Iterable[_core.QgsPointXY]) -> 'QgsGcpTransformerInterface|None': ...
    @staticmethod
    def create(method: 'QgsGcpTransformerInterface.TransformMethod') -> 'QgsGcpTransformerInterface|None': ...
    @staticmethod
    def methodToString(method: 'QgsGcpTransformerInterface.TransformMethod') -> str: ...
    def transform(self, x: float, y: float, inverseTransform: bool = ...) -> typing.Tuple[bool, float, float]: ...
    def method(self) -> 'QgsGcpTransformerInterface.TransformMethod': ...
    def minimumGcpCount(self) -> int: ...
    def updateParametersFromGcps(self, sourceCoordinates: collections.abc.Iterable[_core.QgsPointXY], destinationCoordinates: collections.abc.Iterable[_core.QgsPointXY], invertYAxis: bool = ...) -> bool: ...
    def clone(self) -> 'QgsGcpTransformerInterface|None': ...


class QgsVectorWarper(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, method: QgsGcpTransformerInterface.TransformMethod, points: collections.abc.Iterable[QgsGcpPoint], destinationCrs: _core.QgsCoordinateReferenceSystem) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsVectorWarper') -> None: ...

    def error(self) -> str: ...
    def transformFeatures(self, iterator: _core.QgsFeatureIterator, sink: _core.QgsFeatureSink|None, context: _core.QgsCoordinateTransformContext, feedback: _core.QgsFeedback|None = ...) -> bool: ...


class QgsVectorWarperTask(_core.QgsTask):

    class Result(enum.Enum):
        Success = ... # type: QgsVectorWarperTask.Result
        Canceled = ... # type: QgsVectorWarperTask.Result
        Error = ... # type: QgsVectorWarperTask.Result

    def __init__(self, method: QgsGcpTransformerInterface.TransformMethod, points: collections.abc.Iterable[QgsGcpPoint], destinationCrs: _core.QgsCoordinateReferenceSystem, layer: _core.QgsVectorLayer|None, fileName: str|None) -> None: ...

    def run(self) -> bool: ...
    def errorMessage(self) -> str: ...
    def result(self) -> 'QgsVectorWarperTask.Result': ...
    def cancel(self) -> None: ...


class QgsGridFileWriter(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, interpolator: 'QgsInterpolator|None', outputPath: str|None, extent: _core.QgsRectangle, nCols: int, nRows: int, outputFormat: str|None = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGridFileWriter') -> None: ...

    def noDataValue(self) -> float: ...
    def setNoDataValue(self, noDataValue: float) -> None: ...
    def creationOptions(self) -> list[str]: ...
    def setCreationOptions(self, options: collections.abc.Iterable[str|None]) -> None: ...
    def writeFile(self, feedback: _core.QgsFeedback|None = ...) -> int: ...


class QgsInterpolator(PyQt6.sip.wrapper):

    class Result(enum.IntEnum):
        Success = ... # type: QgsInterpolator.Result
        Canceled = ... # type: QgsInterpolator.Result
        InvalidSource = ... # type: QgsInterpolator.Result
        FeatureGeometryError = ... # type: QgsInterpolator.Result

    class ValueSource(enum.IntEnum):
        Attribute = ... # type: QgsInterpolator.ValueSource
        Z = ... # type: QgsInterpolator.ValueSource
        M = ... # type: QgsInterpolator.ValueSource

    class SourceType(enum.IntEnum):
        Points = ... # type: QgsInterpolator.SourceType
        StructureLines = ... # type: QgsInterpolator.SourceType
        BreakLines = ... # type: QgsInterpolator.SourceType

    class LayerData(PyQt6.sip.wrapper):

        interpolationAttribute = ... # type: int
        source = ... # type: _core.QgsFeatureSource
        sourceType = ... # type: 'QgsInterpolator.SourceType'
        transformContext = ... # type: _core.QgsCoordinateTransformContext
        valueSource = ... # type: 'QgsInterpolator.ValueSource'

        @typing.overload
        def __init__(self) -> None: ...
        @typing.overload
        def __init__(self, a0: 'QgsInterpolator.LayerData') -> None: ...

    @typing.overload
    def __init__(self, layerData: collections.abc.Iterable['QgsInterpolator.LayerData']) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsInterpolator') -> None: ...

    def cacheBaseData(self, feedback: _core.QgsFeedback|None = ...) -> 'QgsInterpolator.Result': ...
    def interpolatePoint(self, x: float, y: float, feedback: _core.QgsFeedback|None = ...) -> typing.Tuple[int, float]: ...


class QgsIDWInterpolator(QgsInterpolator):

    @typing.overload
    def __init__(self, layerData: collections.abc.Iterable[QgsInterpolator.LayerData]) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsIDWInterpolator') -> None: ...

    def distanceCoefficient(self) -> float: ...
    def setDistanceCoefficient(self, coefficient: float) -> None: ...
    def interpolatePoint(self, x: float, y: float, feedback: _core.QgsFeedback|None = ...) -> typing.Tuple[int, float]: ...


class QgsInterpolatorVertexData(PyQt6.sip.wrapper):

    x = ... # type: float
    y = ... # type: float
    z = ... # type: float

    @typing.overload
    def __init__(self, x: float, y: float, z: float) -> None: ...
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsInterpolatorVertexData') -> None: ...


class QgsTinInterpolator(QgsInterpolator):

    class TinInterpolation(enum.IntEnum):
        Linear = ... # type: QgsTinInterpolator.TinInterpolation
        CloughTocher = ... # type: QgsTinInterpolator.TinInterpolation

    @typing.overload
    def __init__(self, inputData: collections.abc.Iterable[QgsInterpolator.LayerData], interpolation: 'QgsTinInterpolator.TinInterpolation' = ..., feedback: _core.QgsFeedback|None = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsTinInterpolator') -> None: ...

    def setTriangulationSink(self, sink: _core.QgsFeatureSink|None) -> None: ...
    @staticmethod
    def triangulationFields() -> _core.QgsFields: ...
    def interpolatePoint(self, x: float, y: float, feedback: _core.QgsFeedback|None) -> typing.Tuple[int, float]: ...


class QgsMeshContours(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, layer: _core.QgsMeshLayer|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsMeshContours') -> None: ...

    def exportPolygons(self, index: _core.QgsMeshDatasetIndex, min_value: float, max_value: float, method: _core.QgsMeshRendererScalarSettings.DataResamplingMethod, feedback: _core.QgsFeedback|None = ...) -> _core.QgsGeometry: ...
    def exportLines(self, index: _core.QgsMeshDatasetIndex, value: float, method: _core.QgsMeshRendererScalarSettings.DataResamplingMethod, feedback: _core.QgsFeedback|None = ...) -> _core.QgsGeometry: ...


class QgsMeshTriangulation(QtCore.QObject):

    def __init__(self) -> None: ...

    def setCrs(self, crs: _core.QgsCoordinateReferenceSystem) -> None: ...
    def triangulatedMesh(self, feedback: _core.QgsFeedback|None = ...) -> _core.QgsMesh: ...
    def addVertex(self, vertex: _core.QgsPoint) -> int: ...
    def addBreakLines(self, lineFeatureIterator: _core.QgsFeatureIterator, valueAttribute: int, transformContext: _core.QgsCoordinateTransform, feedback: _core.QgsFeedback|None = ..., featureCount: int = ...) -> bool: ...
    def addVertices(self, vertexFeatureIterator: _core.QgsFeatureIterator, valueAttribute: int, transform: _core.QgsCoordinateTransform, feedback: _core.QgsFeedback|None = ..., featureCount: int = ...) -> bool: ...


class QgsMeshZValueDatasetGroup(_core.QgsMeshDatasetGroup):

    def __init__(self, datasetGroupName: str|None, mesh: _core.QgsMesh) -> None: ...

    def writeXml(self, doc: QtXml.QDomDocument, context: _core.QgsReadWriteContext) -> QtXml.QDomElement: ...
    def type(self) -> _core.QgsMeshDatasetGroup.Type: ...
    def dataset(self, index: int) -> _core.QgsMeshDataset|None: ...
    def datasetCount(self) -> int: ...
    def datasetMetadata(self, datasetIndex: int) -> _core.QgsMeshDatasetMetadata: ...
    def initialize(self) -> None: ...


class QgsMeshEditingDelaunayTriangulation(_core.QgsMeshAdvancedEditing):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsMeshEditingDelaunayTriangulation') -> None: ...

    def text(self) -> str: ...


class QgsGraphEdge(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGraphEdge') -> None: ...

    def fromVertex(self) -> int: ...
    def toVertex(self) -> int: ...
    def strategies(self) -> list[typing.Any]: ...
    def cost(self, strategyIndex: int) -> typing.Any: ...


class QgsGraphVertex(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, point: _core.QgsPointXY) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGraphVertex') -> None: ...

    def point(self) -> _core.QgsPointXY: ...
    def outgoingEdges(self) -> list[int]: ...
    def incomingEdges(self) -> list[int]: ...


class QgsGraph(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGraph') -> None: ...

    def hasVertex(self, index: int) -> bool: ...
    def hasEdge(self, index: int) -> bool: ...
    def findOppositeEdge(self, index: int) -> int: ...
    def findVertex(self, pt: _core.QgsPointXY) -> int: ...
    def removeEdge(self, index: int) -> None: ...
    def edge(self, idx: int) -> QgsGraphEdge: ...
    def edgeCount(self) -> int: ...
    def removeVertex(self, index: int) -> None: ...
    def vertex(self, idx: int) -> QgsGraphVertex: ...
    def vertexCount(self) -> int: ...
    def addEdge(self, fromVertexIdx: int, toVertexIdx: int, strategies: collections.abc.Iterable[typing.Any]) -> int: ...
    def addVertex(self, pt: _core.QgsPointXY) -> int: ...


class QgsGraphAnalyzer(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGraphAnalyzer') -> None: ...

    @staticmethod
    def shortestTree(source: QgsGraph|None, startVertexIdx: int, criterionNum: int) -> QgsGraph|None: ...
    @staticmethod
    def dijkstra(source: QgsGraph|None, startVertexIdx: int, criterionNum: int, resultTree: collections.abc.Iterable[int] = ..., resultCost: collections.abc.Iterable[float] = ...) -> tuple[list[int], list[float]]: ...


class QgsGraphBuilderInterface(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, crs: _core.QgsCoordinateReferenceSystem, ctfEnabled: bool = ..., topologyTolerance: float = ..., ellipsoidID: str|None = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGraphBuilderInterface') -> None: ...

    def addEdge(self, pt1id: int, pt1: _core.QgsPointXY, pt2id: int, pt2: _core.QgsPointXY, strategies: collections.abc.Iterable[typing.Any]) -> None: ...
    def addVertex(self, id: int, pt: _core.QgsPointXY) -> int: ...
    def distanceArea(self) -> _core.QgsDistanceArea|None: ...
    def topologyTolerance(self) -> float: ...
    def coordinateTransformationEnabled(self) -> bool: ...
    def destinationCrs(self) -> _core.QgsCoordinateReferenceSystem: ...


class QgsGraphBuilder(QgsGraphBuilderInterface):

    def __init__(self, crs: _core.QgsCoordinateReferenceSystem, otfEnabled: bool = ..., topologyTolerance: float = ..., ellipsoidID: str|None = ...) -> None: ...

    def takeGraph(self) -> QgsGraph|None: ...
    def graph(self) -> QgsGraph: ...
    def addEdge(self, pt1id: int, pt1: _core.QgsPointXY, pt2id: int, pt2: _core.QgsPointXY, prop: collections.abc.Iterable[typing.Any]) -> None: ...
    def addVertex(self, id: int, pt: _core.QgsPointXY) -> int: ...


class QgsGraphDirector(QtCore.QObject):

    def __init__(self) -> None: ...

    def name(self) -> str: ...
    def addStrategy(self, prop: 'QgsNetworkStrategy|None') -> None: ...
    def makeGraph(self, builder: QgsGraphBuilderInterface|None, additionalPoints: collections.abc.Iterable[_core.QgsPointXY], feedback: _core.QgsFeedback|None = ...) -> list[_core.QgsPointXY]: ...


class QgsNetworkStrategy(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsNetworkStrategy') -> None: ...

    def cost(self, distance: float, f: _core.QgsFeature) -> typing.Any: ...
    def requiredAttributes(self) -> typing.Any: ...


class QgsNetworkDistanceStrategy(QgsNetworkStrategy):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsNetworkDistanceStrategy') -> None: ...

    def cost(self, distance: float, a1: _core.QgsFeature) -> typing.Any: ...


class QgsNetworkSpeedStrategy(QgsNetworkStrategy):

    @typing.overload
    def __init__(self, attributeId: int, defaultValue: float, toMetricFactor: float) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsNetworkSpeedStrategy') -> None: ...

    def requiredAttributes(self) -> typing.Any: ...
    def cost(self, distance: float, f: _core.QgsFeature) -> typing.Any: ...


class QgsVectorLayerDirector(QgsGraphDirector):

    class Direction(enum.IntEnum):
        DirectionForward = ... # type: QgsVectorLayerDirector.Direction
        DirectionBackward = ... # type: QgsVectorLayerDirector.Direction
        DirectionBoth = ... # type: QgsVectorLayerDirector.Direction

    def __init__(self, source: _core.QgsFeatureSource|None, directionFieldId: int, directDirectionValue: str|None, reverseDirectionValue: str|None, bothDirectionValue: str|None, defaultDirection: 'QgsVectorLayerDirector.Direction') -> None: ...

    def name(self) -> str: ...
    def makeGraph(self, builder: QgsGraphBuilderInterface|None, additionalPoints: collections.abc.Iterable[_core.QgsPointXY], feedback: _core.QgsFeedback|None = ...) -> list[_core.QgsPointXY]: ...


class QgsBatchGeocodeAlgorithm(_core.QgsProcessingFeatureBasedAlgorithm):

    def __init__(self, geocoder: _core.QgsGeocoderInterface|None) -> None: ...

    def outputWkbType(self, inputWkbType: _core.Qgis.WkbType) -> _core.Qgis.WkbType: ...
    def outputFields(self, inputFields: _core.QgsFields) -> _core.QgsFields: ...
    def outputCrs(self, inputCrs: _core.QgsCoordinateReferenceSystem) -> _core.QgsCoordinateReferenceSystem: ...
    def processFeature(self, feature: _core.QgsFeature, a1: _core.QgsProcessingContext, feedback: _core.QgsProcessingFeedback|None) -> list[_core.QgsFeature]: ...
    def prepareAlgorithm(self, parameters: dict[str|None, typing.Any], context: _core.QgsProcessingContext, feedback: _core.QgsProcessingFeedback|None) -> bool: ...
    def outputName(self) -> str: ...
    def supportInPlaceEdit(self, layer: _core.QgsMapLayer|None) -> bool: ...
    def inputLayerTypes(self) -> list[int]: ...
    def groupId(self) -> str: ...
    def group(self) -> str: ...
    def tags(self) -> list[str]: ...
    def initParameters(self, configuration: dict[str|None, typing.Any] = ...) -> None: ...


class QgsNativeAlgorithms(_core.QgsProcessingProvider):

    def __init__(self, parent: QtCore.QObject|None = ...) -> None: ...

    def loadAlgorithms(self) -> None: ...
    def flags(self) -> _core.Qgis.ProcessingProviderFlag: ...
    def supportsNonFileBasedOutput(self) -> bool: ...
    def name(self) -> str: ...
    def helpId(self) -> str: ...
    def id(self) -> str: ...
    def svgIconPath(self) -> str: ...
    def icon(self) -> QtGui.QIcon: ...


class QgsPdalAlgorithms(_core.QgsProcessingProvider):

    def __init__(self, parent: QtCore.QObject|None = ...) -> None: ...

    def loadAlgorithms(self) -> None: ...
    def supportedOutputPointCloudLayerExtensions(self) -> list[str]: ...
    def supportedOutputRasterLayerFormatAndExtensions(self) -> list[tuple[str, str]]: ...
    def supportedOutputVectorLayerExtensions(self) -> list[str]: ...
    def supportsNonFileBasedOutput(self) -> bool: ...
    def name(self) -> str: ...
    def helpId(self) -> str: ...
    def id(self) -> str: ...
    def svgIconPath(self) -> str: ...
    def icon(self) -> QtGui.QIcon: ...


class QgsAlignRaster(PyQt6.sip.wrapper):

    class RasterInfo(PyQt6.sip.wrapper):

        def __init__(self, layerpath: str|None) -> None: ...

        def identify(self, mx: float, my: float) -> float: ...
        def dump(self) -> None: ...
        def origin(self) -> QtCore.QPointF: ...
        def extent(self) -> _core.QgsRectangle: ...
        def gridOffset(self) -> QtCore.QPointF: ...
        def cellSize(self) -> QtCore.QSizeF: ...
        def bandCount(self) -> int: ...
        def rasterSize(self) -> QtCore.QSize: ...
        def crs(self) -> str: ...
        def isValid(self) -> bool: ...

    class ProgressHandler(PyQt6.sip.wrapper):

        @typing.overload
        def __init__(self) -> None: ...
        @typing.overload
        def __init__(self, a0: 'QgsAlignRaster.ProgressHandler') -> None: ...

        def progress(self, complete: float) -> bool: ...

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsAlignRaster') -> None: ...

    @staticmethod
    def suggestedWarpOutput(info: 'QgsAlignRaster.RasterInfo', destWkt: str|None, cellSize: QtCore.QSizeF|None = ..., gridOffset: QtCore.QPointF|None = ..., rect: _core.QgsRectangle|None = ...) -> bool: ...
    def createAndWarp(self, raster: _core.QgsAlignRasterData.RasterItem) -> bool: ...
    def suggestedReferenceLayer(self) -> int: ...
    def dump(self) -> None: ...
    def errorMessage(self) -> str: ...
    def run(self) -> bool: ...
    def alignedRasterExtent(self) -> _core.QgsRectangle: ...
    def alignedRasterSize(self) -> QtCore.QSize: ...
    def checkInputParameters(self) -> bool: ...
    @typing.overload
    def setParametersFromRaster(self, rasterInfo: 'QgsAlignRaster.RasterInfo', customCRSWkt: str|None = ..., customCellSize: QtCore.QSizeF = ..., customGridOffset: QtCore.QPointF = ...) -> bool: ...
    @typing.overload
    def setParametersFromRaster(self, filename: str|None, customCRSWkt: str|None = ..., customCellSize: QtCore.QSizeF = ..., customGridOffset: QtCore.QPointF = ...) -> bool: ...
    def clipExtent(self) -> _core.QgsRectangle: ...
    @typing.overload
    def setClipExtent(self, xmin: float, ymin: float, xmax: float, ymax: float) -> None: ...
    @typing.overload
    def setClipExtent(self, extent: _core.QgsRectangle) -> None: ...
    def destinationCrs(self) -> str: ...
    def setDestinationCrs(self, crsWkt: str|None) -> None: ...
    def cellSize(self) -> QtCore.QSizeF: ...
    @typing.overload
    def setCellSize(self, x: float, y: float) -> None: ...
    @typing.overload
    def setCellSize(self, size: QtCore.QSizeF) -> None: ...
    def gridOffset(self) -> QtCore.QPointF: ...
    def setGridOffset(self, offset: QtCore.QPointF) -> None: ...
    def rasters(self) -> list[_core.QgsAlignRasterData.RasterItem]: ...
    def setRasters(self, list: collections.abc.Iterable[_core.QgsAlignRasterData.RasterItem]) -> None: ...
    def progressHandler(self) -> 'QgsAlignRaster.ProgressHandler|None': ...
    def setProgressHandler(self, progressHandler: 'QgsAlignRaster.ProgressHandler|None') -> None: ...


class QgsNineCellFilter(PyQt6.sip.wrapper):

    class Result(enum.IntEnum):
        Success = ... # type: QgsNineCellFilter.Result
        InputLayerError = ... # type: QgsNineCellFilter.Result
        DriverError = ... # type: QgsNineCellFilter.Result
        CreateOutputError = ... # type: QgsNineCellFilter.Result
        InputBandError = ... # type: QgsNineCellFilter.Result
        OutputBandError = ... # type: QgsNineCellFilter.Result
        RasterSizeError = ... # type: QgsNineCellFilter.Result
        Canceled = ... # type: QgsNineCellFilter.Result

    @typing.overload
    def __init__(self, inputFile: str|None, outputFile: str|None, outputFormat: str|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsNineCellFilter') -> None: ...

    def processNineCellWindow(self) -> typing.Tuple[float, float, float, float, float, float, float, float, float, float]: ...
    def creationOptions(self) -> list[str]: ...
    def setCreationOptions(self, options: collections.abc.Iterable[str|None]) -> None: ...
    def setOutputNodataValue(self, value: float) -> None: ...
    def outputNodataValue(self) -> float: ...
    def setInputNodataValue(self, value: float) -> None: ...
    def inputNodataValue(self) -> float: ...
    def setZFactor(self, factor: float) -> None: ...
    def zFactor(self) -> float: ...
    def setCellSizeY(self, size: float) -> None: ...
    def cellSizeY(self) -> float: ...
    def setCellSizeX(self, size: float) -> None: ...
    def cellSizeX(self) -> float: ...
    def processRaster(self, feedback: _core.QgsFeedback|None = ...) -> 'QgsNineCellFilter.Result': ...


class QgsDerivativeFilter(QgsNineCellFilter):

    @typing.overload
    def __init__(self, inputFile: str|None, outputFile: str|None, outputFormat: str|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsDerivativeFilter') -> None: ...

    def calcFirstDerY(self, x11: float, x21: float, x31: float, x12: float, x22: float, x32: float, x13: float, x23: float, x33: float) -> float: ...
    def calcFirstDerX(self, x11: float, x21: float, x31: float, x12: float, x22: float, x32: float, x13: float, x23: float, x33: float) -> float: ...
    def processNineCellWindow(self) -> typing.Tuple[float, float, float, float, float, float, float, float, float, float]: ...


class QgsAspectFilter(QgsDerivativeFilter):

    @typing.overload
    def __init__(self, inputFile: str|None, outputFile: str|None, outputFormat: str|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsAspectFilter') -> None: ...

    def processNineCellWindow(self) -> typing.Tuple[float, float, float, float, float, float, float, float, float, float]: ...


class QgsHillshadeFilter(QgsDerivativeFilter):

    @typing.overload
    def __init__(self, inputFile: str|None, outputFile: str|None, outputFormat: str|None, lightAzimuth: float = ..., lightAngle: float = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsHillshadeFilter') -> None: ...

    def setLightAngle(self, angle: float) -> None: ...
    def lightAngle(self) -> float: ...
    def setLightAzimuth(self, azimuth: float) -> None: ...
    def lightAzimuth(self) -> float: ...
    def processNineCellWindow(self) -> typing.Tuple[float, float, float, float, float, float, float, float, float, float]: ...


class QgsKernelDensityEstimation(PyQt6.sip.wrapper):

    class Result(enum.IntEnum):
        Success = ... # type: QgsKernelDensityEstimation.Result
        DriverError = ... # type: QgsKernelDensityEstimation.Result
        InvalidParameters = ... # type: QgsKernelDensityEstimation.Result
        FileCreationError = ... # type: QgsKernelDensityEstimation.Result
        RasterIoError = ... # type: QgsKernelDensityEstimation.Result

    class OutputValues(enum.IntEnum):
        Raw = ... # type: QgsKernelDensityEstimation.OutputValues
        Scaled = ... # type: QgsKernelDensityEstimation.OutputValues

    class KernelShape(enum.IntEnum):
        Quartic = ... # type: QgsKernelDensityEstimation.KernelShape
        Triangular = ... # type: QgsKernelDensityEstimation.KernelShape
        Uniform = ... # type: QgsKernelDensityEstimation.KernelShape
        Triweight = ... # type: QgsKernelDensityEstimation.KernelShape
        Epanechnikov = ... # type: QgsKernelDensityEstimation.KernelShape

    class Parameters(PyQt6.sip.wrapper):

        decayRatio = ... # type: float
        outputValues = ... # type: 'QgsKernelDensityEstimation.OutputValues'
        pixelSize = ... # type: float
        radius = ... # type: float
        radiusField = ... # type: str|None
        shape = ... # type: 'QgsKernelDensityEstimation.KernelShape'
        source = ... # type: _core.QgsFeatureSource
        weightField = ... # type: str|None

        @typing.overload
        def __init__(self) -> None: ...
        @typing.overload
        def __init__(self, a0: 'QgsKernelDensityEstimation.Parameters') -> None: ...

    def __init__(self, parameters: 'QgsKernelDensityEstimation.Parameters', outputFile: str|None, outputFormat: str|None) -> None: ...

    def finalise(self) -> 'QgsKernelDensityEstimation.Result': ...
    def addFeature(self, feature: _core.QgsFeature) -> 'QgsKernelDensityEstimation.Result': ...
    def prepare(self) -> 'QgsKernelDensityEstimation.Result': ...
    def run(self) -> 'QgsKernelDensityEstimation.Result': ...


class QgsRasterCalcNode(PyQt6.sip.wrapper):

    class Operator(enum.IntEnum):
        opPLUS = ... # type: QgsRasterCalcNode.Operator
        opMINUS = ... # type: QgsRasterCalcNode.Operator
        opMUL = ... # type: QgsRasterCalcNode.Operator
        opDIV = ... # type: QgsRasterCalcNode.Operator
        opPOW = ... # type: QgsRasterCalcNode.Operator
        opSQRT = ... # type: QgsRasterCalcNode.Operator
        opSIN = ... # type: QgsRasterCalcNode.Operator
        opCOS = ... # type: QgsRasterCalcNode.Operator
        opTAN = ... # type: QgsRasterCalcNode.Operator
        opASIN = ... # type: QgsRasterCalcNode.Operator
        opACOS = ... # type: QgsRasterCalcNode.Operator
        opATAN = ... # type: QgsRasterCalcNode.Operator
        opEQ = ... # type: QgsRasterCalcNode.Operator
        opNE = ... # type: QgsRasterCalcNode.Operator
        opGT = ... # type: QgsRasterCalcNode.Operator
        opLT = ... # type: QgsRasterCalcNode.Operator
        opGE = ... # type: QgsRasterCalcNode.Operator
        opLE = ... # type: QgsRasterCalcNode.Operator
        opAND = ... # type: QgsRasterCalcNode.Operator
        opOR = ... # type: QgsRasterCalcNode.Operator
        opSIGN = ... # type: QgsRasterCalcNode.Operator
        opLOG = ... # type: QgsRasterCalcNode.Operator
        opLOG10 = ... # type: QgsRasterCalcNode.Operator
        opABS = ... # type: QgsRasterCalcNode.Operator
        opMAX = ... # type: QgsRasterCalcNode.Operator
        opMIN = ... # type: QgsRasterCalcNode.Operator
        opNONE = ... # type: QgsRasterCalcNode.Operator

    class Type(enum.IntEnum):
        tOperator = ... # type: QgsRasterCalcNode.Type
        tNumber = ... # type: QgsRasterCalcNode.Type
        tRasterRef = ... # type: QgsRasterCalcNode.Type
        tMatrix = ... # type: QgsRasterCalcNode.Type
        tFunction = ... # type: QgsRasterCalcNode.Type

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, number: float) -> None: ...
    @typing.overload
    def __init__(self, matrix: 'QgsRasterMatrix|None') -> None: ...
    @typing.overload
    def __init__(self, op: 'QgsRasterCalcNode.Operator', left: 'QgsRasterCalcNode|None', right: 'QgsRasterCalcNode|None') -> None: ...
    @typing.overload
    def __init__(self, functionName: str|None, functionArgs: collections.abc.Iterable['QgsRasterCalcNode']) -> None: ...
    @typing.overload
    def __init__(self, rasterName: str|None) -> None: ...

    def cleanRasterReferences(self) -> list[str]: ...
    def referencedLayerNames(self) -> list[str]: ...
    @staticmethod
    def parseRasterCalcString(str: str|None, parserErrorMsg: str|None) -> 'QgsRasterCalcNode|None': ...
    def findNodes(self, type: 'QgsRasterCalcNode.Type') -> list['QgsRasterCalcNode']: ...
    def toString(self, cStyle: bool = ...) -> str: ...
    def setRight(self, right: 'QgsRasterCalcNode|None') -> None: ...
    def setLeft(self, left: 'QgsRasterCalcNode|None') -> None: ...
    def type(self) -> 'QgsRasterCalcNode.Type': ...


class QgsRasterCalculatorEntry(PyQt6.sip.wrapper):

    bandNumber = ... # type: int
    raster = ... # type: _core.QgsRasterLayer
    ref = ... # type: str|None

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsRasterCalculatorEntry') -> None: ...

    @staticmethod
    def rasterEntries() -> list['QgsRasterCalculatorEntry']: ...


class QgsRasterCalculator(PyQt6.sip.wrapper):

    class Result(enum.IntEnum):
        Success = ... # type: QgsRasterCalculator.Result
        CreateOutputError = ... # type: QgsRasterCalculator.Result
        InputLayerError = ... # type: QgsRasterCalculator.Result
        Canceled = ... # type: QgsRasterCalculator.Result
        ParserError = ... # type: QgsRasterCalculator.Result
        MemoryError = ... # type: QgsRasterCalculator.Result
        BandError = ... # type: QgsRasterCalculator.Result
        CalculationError = ... # type: QgsRasterCalculator.Result
        OpenCLKernelBuildError = ... # type: QgsRasterCalculator.Result

    @typing.overload
    def __init__(self, formulaString: str|None, outputFile: str|None, outputFormat: str|None, outputExtent: _core.QgsRectangle, outputCrs: _core.QgsCoordinateReferenceSystem, nOutputColumns: int, nOutputRows: int, rasterEntries: collections.abc.Iterable[QgsRasterCalculatorEntry], transformContext: _core.QgsCoordinateTransformContext) -> None: ...
    @typing.overload
    def __init__(self, formulaString: str|None, outputFile: str|None, outputFormat: str|None, outputExtent: _core.QgsRectangle, nOutputColumns: int, nOutputRows: int, rasterEntries: collections.abc.Iterable[QgsRasterCalculatorEntry], transformContext: _core.QgsCoordinateTransformContext) -> None: ...
    @typing.overload
    @deprecated("""""")
    def __init__(self, formulaString: str|None, outputFile: str|None, outputFormat: str|None, outputExtent: _core.QgsRectangle, outputCrs: _core.QgsCoordinateReferenceSystem, nOutputColumns: int, nOutputRows: int, rasterEntries: collections.abc.Iterable[QgsRasterCalculatorEntry]) -> None: ...
    @typing.overload
    @deprecated("""""")
    def __init__(self, formulaString: str|None, outputFile: str|None, outputFormat: str|None, outputExtent: _core.QgsRectangle, nOutputColumns: int, nOutputRows: int, rasterEntries: collections.abc.Iterable[QgsRasterCalculatorEntry]) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsRasterCalculator') -> None: ...

    def noDataValue(self) -> float: ...
    def setNoDataValue(self, noDataValue: float) -> None: ...
    def creationOptions(self) -> list[str]: ...
    def setCreationOptions(self, options: collections.abc.Iterable[str|None]) -> None: ...
    def lastError(self) -> str: ...
    def processCalculation(self, feedback: _core.QgsFeedback|None = ...) -> 'QgsRasterCalculator.Result': ...


class QgsRasterMatrix(PyQt6.sip.wrapper):

    class OneArgOperator(enum.IntEnum):
        opSQRT = ... # type: QgsRasterMatrix.OneArgOperator
        opSIN = ... # type: QgsRasterMatrix.OneArgOperator
        opCOS = ... # type: QgsRasterMatrix.OneArgOperator
        opTAN = ... # type: QgsRasterMatrix.OneArgOperator
        opASIN = ... # type: QgsRasterMatrix.OneArgOperator
        opACOS = ... # type: QgsRasterMatrix.OneArgOperator
        opATAN = ... # type: QgsRasterMatrix.OneArgOperator
        opSIGN = ... # type: QgsRasterMatrix.OneArgOperator
        opLOG = ... # type: QgsRasterMatrix.OneArgOperator
        opLOG10 = ... # type: QgsRasterMatrix.OneArgOperator
        opABS = ... # type: QgsRasterMatrix.OneArgOperator

    class TwoArgOperator(enum.IntEnum):
        opPLUS = ... # type: QgsRasterMatrix.TwoArgOperator
        opMINUS = ... # type: QgsRasterMatrix.TwoArgOperator
        opMUL = ... # type: QgsRasterMatrix.TwoArgOperator
        opDIV = ... # type: QgsRasterMatrix.TwoArgOperator
        opPOW = ... # type: QgsRasterMatrix.TwoArgOperator
        opEQ = ... # type: QgsRasterMatrix.TwoArgOperator
        opNE = ... # type: QgsRasterMatrix.TwoArgOperator
        opGT = ... # type: QgsRasterMatrix.TwoArgOperator
        opLT = ... # type: QgsRasterMatrix.TwoArgOperator
        opGE = ... # type: QgsRasterMatrix.TwoArgOperator
        opLE = ... # type: QgsRasterMatrix.TwoArgOperator
        opAND = ... # type: QgsRasterMatrix.TwoArgOperator
        opOR = ... # type: QgsRasterMatrix.TwoArgOperator
        opMIN = ... # type: QgsRasterMatrix.TwoArgOperator
        opMAX = ... # type: QgsRasterMatrix.TwoArgOperator

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, m: 'QgsRasterMatrix') -> None: ...

    def absoluteValue(self) -> bool: ...
    def log10(self) -> bool: ...
    def log(self) -> bool: ...
    def changeSign(self) -> bool: ...
    def atangens(self) -> bool: ...
    def tangens(self) -> bool: ...
    def acosinus(self) -> bool: ...
    def cosinus(self) -> bool: ...
    def asinus(self) -> bool: ...
    def sinus(self) -> bool: ...
    def squareRoot(self) -> bool: ...
    def min(self, other: 'QgsRasterMatrix') -> bool: ...
    def max(self, other: 'QgsRasterMatrix') -> bool: ...
    def logicalOr(self, other: 'QgsRasterMatrix') -> bool: ...
    def logicalAnd(self, other: 'QgsRasterMatrix') -> bool: ...
    def lesserEqual(self, other: 'QgsRasterMatrix') -> bool: ...
    def greaterEqual(self, other: 'QgsRasterMatrix') -> bool: ...
    def lesserThan(self, other: 'QgsRasterMatrix') -> bool: ...
    def greaterThan(self, other: 'QgsRasterMatrix') -> bool: ...
    def notEqual(self, other: 'QgsRasterMatrix') -> bool: ...
    def equal(self, other: 'QgsRasterMatrix') -> bool: ...
    def power(self, other: 'QgsRasterMatrix') -> bool: ...
    def divide(self, other: 'QgsRasterMatrix') -> bool: ...
    def multiply(self, other: 'QgsRasterMatrix') -> bool: ...
    def subtract(self, other: 'QgsRasterMatrix') -> bool: ...
    def add(self, other: 'QgsRasterMatrix') -> bool: ...
    def setNodataValue(self, d: float) -> None: ...
    def nodataValue(self) -> float: ...
    def nRows(self) -> int: ...
    def nColumns(self) -> int: ...
    def setData(self, cols: int, rows: int, nodataValue: float) -> float: ...
    def number(self) -> float: ...
    def isNumber(self) -> bool: ...


class QgsRelief(PyQt6.sip.wrapper):

    class Result(enum.IntEnum):
        Success = ... # type: QgsRelief.Result
        InvalidInput = ... # type: QgsRelief.Result
        OutputCreationFailed = ... # type: QgsRelief.Result
        InvalidInputSize = ... # type: QgsRelief.Result
        Canceled = ... # type: QgsRelief.Result

    def __init__(self, inputFile: str|None, outputFile: str|None, outputFormat: str|None) -> None: ...

    def exportFrequencyDistributionToCsv(self, file: str|None) -> bool: ...
    def calculateOptimizedReliefClasses(self) -> list[_core.QgsRasterReliefColor]: ...
    def setReliefColors(self, c: collections.abc.Iterable[_core.QgsRasterReliefColor]) -> None: ...
    def reliefColors(self) -> list[_core.QgsRasterReliefColor]: ...
    def addReliefColorClass(self, color: _core.QgsRasterReliefColor) -> None: ...
    def clearReliefColors(self) -> None: ...
    def setZFactor(self, factor: float) -> None: ...
    def zFactor(self) -> float: ...
    def processRaster(self, feedback: _core.QgsFeedback|None = ...) -> 'QgsRelief.Result': ...


class QgsRuggednessFilter(QgsNineCellFilter):

    @typing.overload
    def __init__(self, inputFile: str|None, outputFile: str|None, outputFormat: str|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsRuggednessFilter') -> None: ...

    def processNineCellWindow(self) -> typing.Tuple[float, float, float, float, float, float, float, float, float, float]: ...


class QgsSlopeFilter(QgsDerivativeFilter):

    @typing.overload
    def __init__(self, inputFile: str|None, outputFile: str|None, outputFormat: str|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsSlopeFilter') -> None: ...

    def processNineCellWindow(self) -> typing.Tuple[float, float, float, float, float, float, float, float, float, float]: ...


class QgsTotalCurvatureFilter(QgsNineCellFilter):

    @typing.overload
    def __init__(self, inputFile: str|None, outputFile: str|None, outputFormat: str|None) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsTotalCurvatureFilter') -> None: ...

    def processNineCellWindow(self) -> typing.Tuple[float, float, float, float, float, float, float, float, float, float]: ...


class QgsFeaturePool(_core.QgsFeatureSink):

    def __init__(self, layer: _core.QgsVectorLayer|None) -> None: ...

    def removeFeature(self, featureId: int) -> None: ...
    def refreshCache(self, feature: _core.QgsFeature, origFeature: _core.QgsFeature) -> None: ...
    def insertFeature(self, feature: _core.QgsFeature, skipLock: bool = ...) -> None: ...
    def layerName(self) -> str: ...
    def crs(self) -> _core.QgsCoordinateReferenceSystem: ...
    def geometryType(self) -> _core.Qgis.GeometryType: ...
    def layerId(self) -> str: ...
    def layer(self) -> _core.QgsVectorLayer|None: ...
    def deleteFeature(self, fid: int) -> None: ...
    def updateFeature(self, feature: _core.QgsFeature) -> None: ...
    def getFeature(self, id: int, feature: _core.QgsFeature) -> bool: ...


class QgsGeometryCheck(PyQt6.sip.wrapper):

    class Flag(enum.IntEnum):
        AvailableInValidation = ... # type: QgsGeometryCheck.Flag

    class Result(enum.IntEnum):
        Success = ... # type: QgsGeometryCheck.Result
        Canceled = ... # type: QgsGeometryCheck.Result
        DuplicatedUniqueId = ... # type: QgsGeometryCheck.Result
        InvalidReferenceLayer = ... # type: QgsGeometryCheck.Result
        GeometryOverlayError = ... # type: QgsGeometryCheck.Result

    class CheckType(enum.IntEnum):
        FeatureNodeCheck = ... # type: QgsGeometryCheck.CheckType
        FeatureCheck = ... # type: QgsGeometryCheck.CheckType
        LayerCheck = ... # type: QgsGeometryCheck.CheckType

    class ChangeType(enum.IntEnum):
        ChangeAdded = ... # type: QgsGeometryCheck.ChangeType
        ChangeRemoved = ... # type: QgsGeometryCheck.ChangeType
        ChangeChanged = ... # type: QgsGeometryCheck.ChangeType

    class ChangeWhat(enum.IntEnum):
        ChangeFeature = ... # type: QgsGeometryCheck.ChangeWhat
        ChangePart = ... # type: QgsGeometryCheck.ChangeWhat
        ChangeRing = ... # type: QgsGeometryCheck.ChangeWhat
        ChangeNode = ... # type: QgsGeometryCheck.ChangeWhat

    class LayerFeatureIds(PyQt6.sip.wrapper):

        @typing.overload
        def __init__(self) -> None: ...
        @typing.overload
        def __init__(self, a0: 'QgsGeometryCheck.LayerFeatureIds') -> None: ...

    class Change(PyQt6.sip.wrapper):

        type = ... # type: 'QgsGeometryCheck.ChangeType'
        vidx = ... # type: _core.QgsVertexId
        what = ... # type: 'QgsGeometryCheck.ChangeWhat'

        @typing.overload
        def __init__(self) -> None: ...
        @typing.overload
        def __init__(self, _what: 'QgsGeometryCheck.ChangeWhat', _type: 'QgsGeometryCheck.ChangeType', _vidx: _core.QgsVertexId = ...) -> None: ...
        @typing.overload
        def __init__(self, a0: 'QgsGeometryCheck.Change') -> None: ...

        def __ne__(self, other: object): ...
        def __eq__(self, other: object): ...

    staticMetaObject = ... # type: QtCore.QMetaObject

    @typing.overload
    def __init__(self, context: 'QgsGeometryCheckContext|None', configuration: dict[str|None, typing.Any]) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGeometryCheck') -> None: ...

    def context(self) -> 'QgsGeometryCheckContext|None': ...
    def checkType(self) -> 'QgsGeometryCheck.CheckType': ...
    def id(self) -> str: ...
    def description(self) -> str: ...
    @deprecated("""""")
    def resolutionMethods(self) -> list[str]: ...
    def availableResolutionMethods(self) -> list['QgsGeometryCheckResolutionMethod']: ...
    def collectErrors(self, featurePools: typing.Any, errors: collections.abc.Iterable['QgsGeometryCheckError'], messages: collections.abc.Iterable[str|None], feedback: _core.QgsFeedback|None, ids: 'QgsGeometryCheck.LayerFeatureIds' = ...) -> typing.Tuple['QgsGeometryCheck.Result', list['QgsGeometryCheckError'], list[str]]: ...
    def flags(self) -> 'QgsGeometryCheck.Flag': ...
    def compatibleGeometryTypes(self) -> list[_core.Qgis.GeometryType]: ...
    def isCompatible(self, layer: _core.QgsVectorLayer|None) -> bool: ...
    def prepare(self, context: 'QgsGeometryCheckContext|None', configuration: dict[str|None, typing.Any]) -> None: ...


class QgsGeometryCheckResolutionMethod(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, id: int, name: str|None, description: str|None, isStable: bool = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGeometryCheckResolutionMethod') -> None: ...

    def description(self) -> str: ...
    def name(self) -> str: ...
    def isStable(self) -> bool: ...
    def id(self) -> int: ...


class QgsGeometryCheckContext(PyQt6.sip.wrapper):

    mapCrs = ... # type: _core.QgsCoordinateReferenceSystem
    reducedTolerance = ... # type: float
    tolerance = ... # type: float
    transformContext = ... # type: _core.QgsCoordinateTransformContext
    uniqueIdFieldIndex = ... # type: int

    def __init__(self, precision: int, mapCrs: _core.QgsCoordinateReferenceSystem, transformContext: _core.QgsCoordinateTransformContext, mProject: _core.QgsProject|None = ..., uniqueIdFieldIndex: int = ...) -> None: ...

    def project(self) -> _core.QgsProject|None: ...


class QgsGeometryCheckError(PyQt6.sip.wrapper):

    class ValueType(enum.IntEnum):
        ValueLength = ... # type: QgsGeometryCheckError.ValueType
        ValueArea = ... # type: QgsGeometryCheckError.ValueType
        ValueOther = ... # type: QgsGeometryCheckError.ValueType

    class Status(enum.IntEnum):
        StatusPending = ... # type: QgsGeometryCheckError.Status
        StatusFixFailed = ... # type: QgsGeometryCheckError.Status
        StatusFixed = ... # type: QgsGeometryCheckError.Status
        StatusObsolete = ... # type: QgsGeometryCheckError.Status

    @typing.overload
    def __init__(self, check: QgsGeometryCheck|None, layerFeature: 'QgsGeometryCheckerUtils.LayerFeature', errorLocation: _core.QgsPointXY, vidx: _core.QgsVertexId = ..., value: typing.Any = ..., valueType: 'QgsGeometryCheckError.ValueType' = ...) -> None: ...
    @typing.overload
    def __init__(self, check: QgsGeometryCheck|None, layerId: str|None, featureId: int, geometry: _core.QgsGeometry, errorLocation: _core.QgsPointXY, vidx: _core.QgsVertexId = ..., value: typing.Any = ..., valueType: 'QgsGeometryCheckError.ValueType' = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGeometryCheckError') -> None: ...

    def icon(self) -> QtGui.QIcon: ...
    def update(self, other: 'QgsGeometryCheckError|None') -> None: ...
    def closeMatch(self, a0: 'QgsGeometryCheckError|None') -> bool: ...
    def isEqual(self, other: 'QgsGeometryCheckError|None') -> bool: ...
    def setObsolete(self) -> None: ...
    def setFixFailed(self, reason: str|None) -> None: ...
    def setFixed(self, method: int) -> None: ...
    def resolutionMessage(self) -> str: ...
    def status(self) -> 'QgsGeometryCheckError.Status': ...
    def vidx(self) -> _core.QgsVertexId: ...
    def valueType(self) -> 'QgsGeometryCheckError.ValueType': ...
    def value(self) -> typing.Any: ...
    def location(self) -> _core.QgsPointXY: ...
    def description(self) -> str: ...
    def affectedAreaBBox(self) -> _core.QgsRectangle: ...
    def contextBoundingBox(self) -> _core.QgsRectangle: ...
    def geometry(self) -> _core.QgsGeometry: ...
    def featureId(self) -> int: ...
    def layerId(self) -> str: ...
    def check(self) -> QgsGeometryCheck|None: ...


class QgsGeometryCheckerUtils(PyQt6.sip.wrapper):

    class LayerFeature(PyQt6.sip.wrapper):

        @typing.overload
        def __init__(self, pool: QgsFeaturePool|None, feature: _core.QgsFeature, context: QgsGeometryCheckContext|None, useMapCrs: bool) -> None: ...
        @typing.overload
        def __init__(self, a0: 'QgsGeometryCheckerUtils.LayerFeature') -> None: ...

        def useMapCrs(self) -> bool: ...
        def __ne__(self, other: object): ...
        def __eq__(self, other: object): ...
        def id(self) -> str: ...
        def geometry(self) -> _core.QgsGeometry: ...
        def layerId(self) -> str: ...
        def feature(self) -> _core.QgsFeature: ...

    class LayerFeatures(PyQt6.sip.wrapper):

        def __init__(self, a0: 'QgsGeometryCheckerUtils.LayerFeatures') -> None: ...

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGeometryCheckerUtils') -> None: ...


class QgsGeometryCheckFactory(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGeometryCheckFactory') -> None: ...

    def checkType(self) -> QgsGeometryCheck.CheckType: ...
    def flags(self) -> QgsGeometryCheck.Flag: ...
    def isCompatible(self, layer: _core.QgsVectorLayer|None) -> bool: ...
    def description(self) -> str: ...
    def id(self) -> str: ...
    def createGeometryCheck(self, context: QgsGeometryCheckContext|None, configuration: dict[str|None, typing.Any]) -> QgsGeometryCheck|None: ...


class QgsGeometryCheckRegistry(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGeometryCheckRegistry') -> None: ...

    def registerGeometryCheck(self, checkFactory: QgsGeometryCheckFactory|None) -> None: ...
    def geometryCheckFactories(self, layer: _core.QgsVectorLayer|None, type: QgsGeometryCheck.CheckType, flags: QgsGeometryCheck.Flag = ...) -> list[QgsGeometryCheckFactory]: ...
    def geometryCheck(self, checkId: str|None, context: QgsGeometryCheckContext|None, geometryCheckConfig: dict[str|None, typing.Any]) -> QgsGeometryCheck|None: ...


class QgsSingleGeometryCheckError(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, check: 'QgsSingleGeometryCheck|None', geometry: _core.QgsGeometry, errorLocation: _core.QgsGeometry, vertexId: _core.QgsVertexId = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsSingleGeometryCheckError') -> None: ...

    def vertexId(self) -> _core.QgsVertexId: ...
    def errorLocation(self) -> _core.QgsGeometry: ...
    def check(self) -> 'QgsSingleGeometryCheck|None': ...
    def description(self) -> str: ...
    def isEqual(self, other: 'QgsSingleGeometryCheckError|None') -> bool: ...
    def update(self, other: 'QgsSingleGeometryCheckError|None') -> None: ...


class QgsGeometryCheckErrorSingle(QgsGeometryCheckError):

    @typing.overload
    def __init__(self, singleError: QgsSingleGeometryCheckError|None, layerFeature: QgsGeometryCheckerUtils.LayerFeature) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGeometryCheckErrorSingle') -> None: ...

    def singleError(self) -> QgsSingleGeometryCheckError|None: ...


class QgsSingleGeometryCheck(QgsGeometryCheck):

    @typing.overload
    def __init__(self, context: QgsGeometryCheckContext|None, configuration: dict[str|None, typing.Any]) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsSingleGeometryCheck') -> None: ...

    def processGeometry(self, geometry: _core.QgsGeometry) -> list[QgsSingleGeometryCheckError]: ...
    def collectErrors(self, featurePools: typing.Any, errors: collections.abc.Iterable[QgsGeometryCheckError], messages: collections.abc.Iterable[str|None], feedback: _core.QgsFeedback|None = ..., ids: QgsGeometryCheck.LayerFeatureIds = ...) -> QgsGeometryCheck.Result: ...


class QgsGeometrySnapper(QtCore.QObject):

    class SnapMode(enum.IntEnum):
        PreferNodes = ... # type: QgsGeometrySnapper.SnapMode
        PreferClosest = ... # type: QgsGeometrySnapper.SnapMode
        PreferNodesNoExtraVertices = ... # type: QgsGeometrySnapper.SnapMode
        PreferClosestNoExtraVertices = ... # type: QgsGeometrySnapper.SnapMode
        EndPointPreferNodes = ... # type: QgsGeometrySnapper.SnapMode
        EndPointPreferClosest = ... # type: QgsGeometrySnapper.SnapMode
        EndPointToEndPoint = ... # type: QgsGeometrySnapper.SnapMode

    def __init__(self, referenceSource: _core.QgsFeatureSource|None) -> None: ...

    featureSnapped: typing.ClassVar[QtCore.pyqtSignal]
    def snapFeatures(self, features: collections.abc.Iterable[_core.QgsFeature], snapTolerance: float, mode: 'QgsGeometrySnapper.SnapMode' = ...) -> list[_core.QgsFeature]: ...
    @typing.overload
    def snapGeometry(self, geometry: _core.QgsGeometry, snapTolerance: float, mode: 'QgsGeometrySnapper.SnapMode' = ...) -> _core.QgsGeometry: ...
    @typing.overload
    @staticmethod
    def snapGeometry(geometry: _core.QgsGeometry, snapTolerance: float, referenceGeometries: collections.abc.Iterable[_core.QgsGeometry], mode: 'QgsGeometrySnapper.SnapMode' = ...) -> _core.QgsGeometry: ...


class QgsInternalGeometrySnapper(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, snapTolerance: float, mode: QgsGeometrySnapper.SnapMode = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsInternalGeometrySnapper') -> None: ...

    def snappedGeometries(self) -> typing.Any: ...
    def snapFeature(self, feature: _core.QgsFeature) -> _core.QgsGeometry: ...


class QgsGeometrySnapperSingleSource(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsGeometrySnapperSingleSource') -> None: ...

    @staticmethod
    def run(source: _core.QgsFeatureSource, sink: _core.QgsFeatureSink, thresh: float, feedback: _core.QgsFeedback|None) -> int: ...


class QgsZonalStatistics(PyQt6.sip.wrapper):

    @typing.overload
    def __init__(self, polygonLayer: _core.QgsVectorLayer|None, rasterLayer: _core.QgsRasterLayer|None, attributePrefix: str|None = ..., rasterBand: int = ..., stats: _core.Qgis.ZonalStatistic = ...) -> None: ...
    @typing.overload
    def __init__(self, polygonLayer: _core.QgsVectorLayer|None, rasterInterface: _core.QgsRasterInterface|None, rasterCrs: _core.QgsCoordinateReferenceSystem, rasterUnitsPerPixelX: float, rasterUnitsPerPixelY: float, attributePrefix: str|None = ..., rasterBand: int = ..., stats: _core.Qgis.ZonalStatistic = ...) -> None: ...
    @typing.overload
    def __init__(self, a0: 'QgsZonalStatistics') -> None: ...

    @staticmethod
    def shortName(statistic: _core.Qgis.ZonalStatistic) -> str: ...
    @staticmethod
    def displayName(statistic: _core.Qgis.ZonalStatistic) -> str: ...
    @typing.overload
    def calculateStatistics(self, feedback: _core.QgsFeedback|None) -> _core.Qgis.ZonalStatisticResult: ...
    @typing.overload
    @staticmethod
    def calculateStatistics(rasterInterface: _core.QgsRasterInterface|None, geometry: _core.QgsGeometry, cellSizeX: float, cellSizeY: float, rasterBand: int, statistics: _core.Qgis.ZonalStatistic) -> dict[int, typing.Any]: ...



"""
***************************************************************************
    MultipleInputDialog.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os
import warnings
from pathlib import Path

from qgis.core import (
    QgsProcessing,
    QgsProcessingModelChildParameterSource,
    QgsProviderRegistry,
    QgsSettings,
    QgsVectorFileWriter,
)
from qgis.PyQt import uic
from qgis.PyQt.QtCore import QByteArray, QCoreApplication, Qt
from qgis.PyQt.QtGui import QStandardItem, QStandardItemModel
from qgis.PyQt.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QPushButton,
)

pluginPath = os.path.split(os.path.dirname(__file__))[0]
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    WIDGET, BASE = uic.loadUiType(
        os.path.join(pluginPath, "ui", "DlgMultipleSelection.ui")
    )


class MultipleInputDialog(BASE, WIDGET):
    def __init__(self, options, selectedoptions=None, datatype=None):
        super().__init__(None)
        self.setupUi(self)
        self.datatype = datatype
        self.model = None

        self.options = []
        for i, option in enumerate(options):
            if option is None or isinstance(option, str):
                self.options.append((i, option))
            else:
                self.options.append((option[0], option[1]))

        self.selectedoptions = selectedoptions or []

        # Additional buttons
        self.btnSelectAll = QPushButton(self.tr("Select All"))
        self.buttonBox.addButton(
            self.btnSelectAll, QDialogButtonBox.ButtonRole.ActionRole
        )
        self.btnClearSelection = QPushButton(self.tr("Clear Selection"))
        self.buttonBox.addButton(
            self.btnClearSelection, QDialogButtonBox.ButtonRole.ActionRole
        )
        self.btnToggleSelection = QPushButton(self.tr("Toggle Selection"))
        self.buttonBox.addButton(
            self.btnToggleSelection, QDialogButtonBox.ButtonRole.ActionRole
        )
        if self.datatype is not None:
            btnAddFile = QPushButton(
                QCoreApplication.translate("MultipleInputDialog", "Add File(s)…")
            )
            btnAddFile.clicked.connect(self.addFiles)
            self.buttonBox.addButton(btnAddFile, QDialogButtonBox.ButtonRole.ActionRole)

            btnAddDir = QPushButton(
                QCoreApplication.translate("MultipleInputDialog", "Add Directory…")
            )
            btnAddDir.clicked.connect(self.addDirectory)
            self.buttonBox.addButton(btnAddDir, QDialogButtonBox.ButtonRole.ActionRole)

        self.btnSelectAll.clicked.connect(lambda: self.selectAll(True))
        self.btnClearSelection.clicked.connect(lambda: self.selectAll(False))
        self.btnToggleSelection.clicked.connect(self.toggleSelection)

        self.settings = QgsSettings()
        self.restoreGeometry(
            self.settings.value("/Processing/multipleInputDialogGeometry", QByteArray())
        )

        self.lstLayers.setSelectionMode(
            QAbstractItemView.SelectionMode.ExtendedSelection
        )
        self.lstLayers.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)

        self.populateList()
        self.finished.connect(self.saveWindowGeometry)

    def saveWindowGeometry(self):
        self.settings.setValue(
            "/Processing/multipleInputDialogGeometry", self.saveGeometry()
        )

    def populateList(self):
        self.model = QStandardItemModel()
        for value, text in self.options:
            item = QStandardItem(text)
            item.setData(value, Qt.ItemDataRole.UserRole)
            item.setCheckState(
                Qt.CheckState.Checked
                if value in self.selectedoptions
                else Qt.CheckState.Unchecked
            )
            item.setCheckable(True)
            item.setDropEnabled(False)
            self.model.appendRow(item)

        # add extra options (e.g. manually added layers)
        for t in [o for o in self.selectedoptions if not isinstance(o, int)]:
            if isinstance(t, QgsProcessingModelChildParameterSource):
                item = QStandardItem(t.staticValue())
            else:
                item = QStandardItem(t)
            item.setData(item.text(), Qt.ItemDataRole.UserRole)
            item.setCheckState(Qt.CheckState.Checked)
            item.setCheckable(True)
            item.setDropEnabled(False)
            self.model.appendRow(item)

        self.lstLayers.setModel(self.model)

    def accept(self):
        self.selectedoptions = []
        model = self.lstLayers.model()
        for i in range(model.rowCount()):
            item = model.item(i)
            if item.checkState() == Qt.CheckState.Checked:
                self.selectedoptions.append(item.data(Qt.ItemDataRole.UserRole))
        QDialog.accept(self)

    def reject(self):
        self.selectedoptions = None
        QDialog.reject(self)

    def getItemsToModify(self):
        items = []
        if len(self.lstLayers.selectedIndexes()) > 1:
            for i in self.lstLayers.selectedIndexes():
                items.append(self.model.itemFromIndex(i))
        else:
            for i in range(self.model.rowCount()):
                items.append(self.model.item(i))
        return items

    def selectAll(self, value):
        for item in self.getItemsToModify():
            item.setCheckState(
                Qt.CheckState.Checked if value else Qt.CheckState.Unchecked
            )

    def toggleSelection(self):
        for item in self.getItemsToModify():
            checked = item.checkState() == Qt.CheckState.Checked
            item.setCheckState(
                Qt.CheckState.Unchecked if checked else Qt.CheckState.Checked
            )

    def getFileFilter(self, datatype):
        """
        Returns a suitable file filter pattern for the specified parameter definition
        :param param:
        :return:
        """
        if datatype == QgsProcessing.SourceType.TypeRaster:
            return QgsProviderRegistry.instance().fileRasterFilters()
        elif datatype == QgsProcessing.SourceType.TypeFile:
            return self.tr("All files (*.*)")
        else:
            exts = QgsVectorFileWriter.supportedFormatExtensions()
            for i in range(len(exts)):
                exts[i] = self.tr("{0} files (*.{1})").format(
                    exts[i].upper(), exts[i].lower()
                )
            return self.tr("All files (*.*)") + ";;" + ";;".join(exts)

    def addFiles(self):
        filter = self.getFileFilter(self.datatype)

        settings = QgsSettings()
        path = str(settings.value("/Processing/LastInputPath"))

        ret, selected_filter = QFileDialog.getOpenFileNames(
            self, self.tr("Select File(s)"), path, filter
        )
        if ret:
            files = list(ret)
            settings.setValue(
                "/Processing/LastInputPath", os.path.dirname(str(files[0]))
            )
            for filename in files:
                item = QStandardItem(filename)
                item.setData(filename, Qt.ItemDataRole.UserRole)
                item.setCheckState(Qt.CheckState.Checked)
                item.setCheckable(True)
                item.setDropEnabled(False)
                self.model.appendRow(item)

    def addDirectory(self):
        settings = QgsSettings()
        path = str(settings.value("/Processing/LastInputPath"))

        ret = QFileDialog.getExistingDirectory(self, self.tr("Select File(s)"), path)
        if ret:
            exts = []

            if self.datatype == QgsProcessing.SourceType.TypeVector:
                exts = QgsVectorFileWriter.supportedFormatExtensions()
            elif self.datatype == QgsProcessing.SourceType.TypeRaster:
                for t in (
                    QgsProviderRegistry.instance().fileRasterFilters().split(";;")[1:]
                ):
                    for e in t[t.index("(") + 1 : -1].split(" "):
                        if e != "*.*" and e.startswith("*."):
                            exts.append(e[2:])

            files = []
            for pp in Path(ret).rglob("*"):
                if not pp.is_file():
                    continue

                if exts and pp.suffix[1:] not in exts:
                    continue

                p = pp.as_posix()

                files.append(p)

            settings.setValue("/Processing/LastInputPath", ret)

            for filename in files:
                item = QStandardItem(filename)
                item.setData(filename, Qt.ItemDataRole.UserRole)
                item.setCheckState(Qt.CheckState.Checked)
                item.setCheckable(True)
                item.setDropEnabled(False)
                self.model.appendRow(item)



"""
***************************************************************************
    QgisAlgorithmTests.py
    ---------------------
    Date                 : January 2019
    Copyright            : (C) 2019 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nyall Dawson"
__date__ = "January 2019"
__copyright__ = "(C) 2019, Nyall Dawson"

import gc
import shutil
import unittest

import nose2
from qgis.analysis import QgsNativeAlgorithms
from qgis.core import (
    QgsApplication,
    QgsProcessing,
    QgsProcessingContext,
    QgsProject,
    QgsVectorLayer,
)
from qgis.PyQt import sip
from qgis.testing import QgisTestCase, start_app

import processing
from processing.tests.TestData import points


class TestProcessingGeneral(QgisTestCase):
    @classmethod
    def setUpClass(cls):
        start_app()
        from processing.core.Processing import Processing

        Processing.initialize()
        cls.cleanup_paths = []
        cls.in_place_layers = {}
        cls.vector_layer_params = {}

    @classmethod
    def tearDownClass(cls):
        from processing.core.Processing import Processing

        Processing.deinitialize()
        for path in cls.cleanup_paths:
            shutil.rmtree(path)

    def testRun(self):
        context = QgsProcessingContext()

        # try running an alg using processing.run - ownership of result layer should be transferred back to the caller
        res = processing.run(
            "qgis:buffer",
            {
                "DISTANCE": 1,
                "INPUT": points(),
                "OUTPUT": QgsProcessing.TEMPORARY_OUTPUT,
            },
            context=context,
        )
        self.assertIn("OUTPUT", res)
        # output should be the layer instance itself
        self.assertIsInstance(res["OUTPUT"], QgsVectorLayer)
        # Python should have ownership
        self.assertTrue(sip.ispyowned(res["OUTPUT"]))
        del context
        gc.collect()
        self.assertFalse(sip.isdeleted(res["OUTPUT"]))

        # now try using processing.run with is_child_algorithm = True. Ownership should remain with the context
        context = QgsProcessingContext()
        res = processing.run(
            "qgis:buffer",
            {
                "DISTANCE": 1,
                "INPUT": points(),
                "OUTPUT": QgsProcessing.TEMPORARY_OUTPUT,
            },
            context=context,
            is_child_algorithm=True,
        )
        self.assertIn("OUTPUT", res)
        # output should be a layer string reference, NOT the layer itself
        self.assertIsInstance(res["OUTPUT"], str)
        layer = context.temporaryLayerStore().mapLayer(res["OUTPUT"])
        self.assertIsInstance(layer, QgsVectorLayer)
        # context should have ownership
        self.assertFalse(sip.ispyowned(layer))
        del context
        gc.collect()
        self.assertTrue(sip.isdeleted(layer))

    def testRunAndLoadResults(self):
        QgsProject.instance().removeAllMapLayers()
        context = QgsProcessingContext()

        # try running an alg using processing.runAndLoadResults - ownership of result layer should be transferred to
        # project, and layer should be present in project
        res = processing.runAndLoadResults(
            "qgis:buffer",
            {
                "DISTANCE": 1,
                "INPUT": points(),
                "OUTPUT": QgsProcessing.TEMPORARY_OUTPUT,
            },
            context=context,
        )
        self.assertIn("OUTPUT", res)
        # output should be the layer path
        self.assertIsInstance(res["OUTPUT"], str)

        self.assertEqual(
            context.layersToLoadOnCompletion()[res["OUTPUT"]].project,
            QgsProject.instance(),
        )
        layer = QgsProject.instance().mapLayer(res["OUTPUT"])
        self.assertIsInstance(layer, QgsVectorLayer)

        # Python should NOT have ownership
        self.assertFalse(sip.ispyowned(layer))

    def testProviders(self):
        """
        When run from a standalone script (like this test), ensure that the providers from separate plugins are available
        """
        providers = [p.id() for p in QgsApplication.processingRegistry().providers()]
        self.assertIn("qgis", providers)
        self.assertIn("native", providers)
        self.assertIn("gdal", providers)
        self.assertIn("project", providers)
        self.assertIn("script", providers)
        self.assertIn("model", providers)
        self.assertIn("grass", providers)


if __name__ == "__main__":
    nose2.main()



"""Plugin dependencies selection dialog

.. note:: This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

"""

__author__ = "elpaso@itopen.it"
__date__ = "2018-09-19"
__copyright__ = "Copyright 2018, GISCE-TI S.L."

from pathlib import Path

from qgis.PyQt import QtCore, QtWidgets, uic

Ui_QgsPluginDependenciesDialogBase, _ = uic.loadUiType(
    Path(__file__).parent / "qgsplugindependenciesdialogbase.ui"
)


class QgsPluginDependenciesDialog(
    QtWidgets.QDialog, Ui_QgsPluginDependenciesDialogBase
):
    """A dialog that shows plugin dependencies and offers a way to install or upgrade the
    dependencies.
    """

    def __init__(self, plugin_name, to_install, to_upgrade, not_found, parent=None):
        """Creates the dependencies dialog

        :param plugin_name: the name of the parent plugin
        :type plugin_name: str
        :param to_install: dict of plugins that need to be installed, keyed by plugin name
        :type to_install: dict
        :param to_upgrade: dict of plugins that need to be upgraded, keyed by plugin name
        :type to_upgrade: dict
        :param not_found: dict of plugins that were not found (unavailable), keyed by plugin name
        :type not_found: dict
        :param parent: parent object, defaults to None
        :param parent: QWidget, optional
        """

        super().__init__(parent)
        self.setupUi(self)
        self.setWindowTitle(self.tr("Plugin Dependencies Manager"))
        self.mPluginDependenciesLabel.setText(
            self.tr("Plugin dependencies for <b>%s</b>") % plugin_name
        )
        self.setStyleSheet("QTableView { padding: 20px;}")
        # Name, Version Installed, Version Required, Version Available, Action Checkbox
        self.pluginList.setColumnCount(5)
        self.pluginList.setHorizontalHeaderLabels(
            [
                self.tr("Name"),
                self.tr("Installed"),
                self.tr("Required"),
                self.tr("Available"),
                self.tr("Action"),
            ]
        )
        self.pluginList.setRowCount(len(not_found) + len(to_install) + len(to_upgrade))
        self.__actions = {}

        def _display(txt):
            if txt is None:
                return ""
            return txt

        def _make_row(data, i, name):
            widget = QtWidgets.QLabel("<b>%s</b>" % name)
            widget.p_id = data["id"]
            widget.action = data["action"]
            widget.use_stable_version = data["use_stable_version"]
            self.pluginList.setCellWidget(i, 0, widget)
            self.pluginList.resizeColumnToContents(0)
            widget = QtWidgets.QTableWidgetItem(_display(data["version_installed"]))
            widget.setTextAlignment(
                QtCore.Qt.AlignmentFlag.AlignHCenter
                | QtCore.Qt.AlignmentFlag.AlignVCenter
            )
            self.pluginList.setItem(i, 1, widget)
            widget = QtWidgets.QTableWidgetItem(_display(data["version_required"]))
            widget.setTextAlignment(
                QtCore.Qt.AlignmentFlag.AlignHCenter
                | QtCore.Qt.AlignmentFlag.AlignVCenter
            )
            self.pluginList.setItem(i, 2, widget)
            widget = QtWidgets.QTableWidgetItem(_display(data["version_available"]))
            widget.setTextAlignment(
                QtCore.Qt.AlignmentFlag.AlignHCenter
                | QtCore.Qt.AlignmentFlag.AlignVCenter
            )
            self.pluginList.setItem(i, 3, widget)

        i = 0
        for name, data in to_install.items():
            _make_row(data, i, name)
            widget = QtWidgets.QCheckBox(self.tr("Install"))
            widget.setChecked(True)
            self.pluginList.setCellWidget(i, 4, widget)
            i += 1

        for name, data in to_upgrade.items():
            _make_row(data, i, name)
            widget = QtWidgets.QCheckBox(self.tr("Upgrade"))
            widget.setChecked(True)
            self.pluginList.setCellWidget(i, 4, widget)
            i += 1

        for name, data in not_found.items():
            _make_row(data, i, name)
            widget = QtWidgets.QLabel(self.tr("Fix manually"))
            self.pluginList.setCellWidget(i, 4, widget)
            i += 1

    def actions(self):
        """Returns the list of actions

        :return: dict of actions
        :rtype: dict
        """

        return self.__actions

    def accept(self):
        self.__actions = {}
        for i in range(self.pluginList.rowCount()):
            try:
                if self.pluginList.cellWidget(i, 4).isChecked():
                    self.__actions[self.pluginList.cellWidget(i, 0).p_id] = {
                        "action": self.pluginList.cellWidget(i, 0).action,
                        "use_stable_version": self.pluginList.cellWidget(
                            i, 0
                        ).use_stable_version,
                    }
            except:
                pass
        super().accept()



"""
***************************************************************************
    v_extrude.py
    ------------
    Date                 : March 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "March 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

from qgis.PyQt.QtCore import QCoreApplication


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    """Verify if we have the right parameters"""
    height = alg.parameterAsDouble(parameters, "height", context)
    height_column = alg.parameterAsString(parameters, "height_column", context)
    if (height and height_column) or (not height and not height_column):
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "You need to set either a fixed height value or the height column!",
        )

    return True, None



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsDataSourceUri,
    QgsMapLayerType,
    QgsMessageOutput,
    QgsProject,
)
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QApplication, QPushButton


class DBManagerPlugin:
    def __init__(self, iface):
        self.iface = iface
        self.dlg = None

    def initGui(self):
        self.action = QAction(
            QgsApplication.getThemeIcon("dbmanager.svg"),
            QApplication.translate("DBManagerPlugin", "DB Manager…"),
            self.iface.mainWindow(),
        )

        self.action.setObjectName("dbManager")
        self.action.triggered.connect(self.run)
        # Add toolbar button and menu item
        if hasattr(self.iface, "addDatabaseToolBarIcon"):
            self.iface.addDatabaseToolBarIcon(self.action)
        else:
            self.iface.addToolBarIcon(self.action)
        if hasattr(self.iface, "addPluginToDatabaseMenu"):
            self.iface.addPluginToDatabaseMenu(
                QApplication.translate("DBManagerPlugin", None), self.action
            )
        else:
            self.iface.addPluginToMenu(
                QApplication.translate("DBManagerPlugin", "DB Manager"), self.action
            )

        def install_community(_):
            self.iface.showPluginManager(
                tabIndex=0, searchTerm=r"DB Manager \(community\)"
            )

        if self.iface.messageBar() is not None:
            message_widget = self.iface.messageBar().createMessage(
                QApplication.translate("DBManagerPlugin", "DB Manager"),
                QApplication.translate(
                    "DBManagerPlugin",
                    'The DB Manager plugin will be removed in a future QGIS release. Please install the "DB Manager (community)" replacement plugin instead.',
                ),
            )
            install_button = QPushButton("Install Now")
            install_button.clicked.connect(install_community)
            message_widget.layout().addWidget(install_button)
            self.iface.messageBar().pushWidget(
                message_widget, Qgis.MessageLevel.Warning, 0
            )

    def unload(self):
        # Remove the plugin menu item and icon
        if hasattr(self.iface, "databaseMenu"):
            self.iface.databaseMenu().removeAction(self.action)
        else:
            self.iface.removePluginMenu(
                QApplication.translate("DBManagerPlugin", "DB Manager"), self.action
            )
        if hasattr(self.iface, "removeDatabaseToolBarIcon"):
            self.iface.removeDatabaseToolBarIcon(self.action)
        else:
            self.iface.removeToolBarIcon(self.action)

        if self.dlg is not None:
            self.dlg.close()

    def run(self):
        # keep opened only one instance
        if self.dlg is None:
            from .db_manager import DBManager

            self.dlg = DBManager(self.iface)
            self.dlg.destroyed.connect(self.onDestroyed)
        self.dlg.show()
        self.dlg.raise_()
        self.dlg.setWindowState(
            self.dlg.windowState() & ~Qt.WindowState.WindowMinimized
        )
        self.dlg.activateWindow()

    def onDestroyed(self, obj):
        self.dlg = None



"""
***************************************************************************
    v_net_connectivity.py
    ---------------------
    Date                 : December 2015
    Copyright            : (C) 2015 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "December 2015"
__copyright__ = "(C) 2015, Médéric Ribreux"

from qgis.PyQt.QtCore import QCoreApplication

from .v_net import incorporatePoints, variableOutput


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    """Verify if we have the right parameters"""
    params = ["where", "cats"]
    values = []
    for param in params:
        for i in range(1, 3):
            values.append(alg.parameterAsString(parameters, f"set{i}_{param}", context))

    if (values[0] or values[2]) and (values[1] or values[3]):
        return True, None

    return False, QCoreApplication.translate(
        "GrassAlgorithmExt",
        "You need to set at least setX_where or setX_cats parameters for each set!",
    )


def processCommand(alg, parameters, context, feedback):
    incorporatePoints(alg, parameters, context, feedback)


def processOutputs(alg, parameters, context, feedback):
    outputParameter = {"output": ["output", "point", 2, True]}
    variableOutput(alg, outputParameter, parameters, context)



"""
***************************************************************************
    FileSelectionPanel.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os
import warnings

from qgis.core import QgsSettings
from qgis.PyQt import uic
from qgis.PyQt.QtWidgets import QFileDialog

from processing.tools.system import isWindows

pluginPath = os.path.split(os.path.dirname(__file__))[0]

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    WIDGET, BASE = uic.loadUiType(
        os.path.join(pluginPath, "ui", "widgetBaseSelector.ui")
    )


class FileSelectionPanel(BASE, WIDGET):
    def __init__(self, isFolder, ext=None):
        super().__init__(None)
        self.setupUi(self)

        self.ext = ext or "*"
        self.isFolder = isFolder

        self.btnSelect.clicked.connect(self.showSelectionDialog)

    def showSelectionDialog(self):
        # Find the file dialog's working directory
        settings = QgsSettings()
        text = self.leText.text()
        if os.path.isdir(text):
            path = text
        elif os.path.isdir(os.path.dirname(text)):
            path = os.path.dirname(text)
        elif settings.contains("/Processing/LastInputPath"):
            path = settings.value("/Processing/LastInputPath")
        else:
            path = ""

        if self.isFolder:
            folder = QFileDialog.getExistingDirectory(
                self, self.tr("Select Folder"), path
            )
            if folder:
                self.leText.setText(folder)
                settings.setValue("/Processing/LastInputPath", os.path.dirname(folder))
        else:
            filenames, selected_filter = QFileDialog.getOpenFileNames(
                self,
                self.tr("Select File"),
                path,
                self.tr("{} files").format(self.ext.upper())
                + " (*."
                + self.ext
                + self.tr(");;All files (*.*)"),
            )
            if filenames:
                self.leText.setText(";".join(filenames))
                settings.setValue(
                    "/Processing/LastInputPath", os.path.dirname(filenames[0])
                )

    def getValue(self):
        s = self.leText.text()
        if isWindows():
            s = s.replace("\\", "/")
        return s

    def setText(self, text):
        self.leText.setText(text)



"""
***************************************************************************
    plugin_test.py
    ---------------------
    Date                 : May 2017
    Copyright            : (C) 2017, Sandro Santilli
    Email                : strk at kbt dot io
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Sandro Santilli"
__date__ = "May 2017"
__copyright__ = "(C) 2017, Sandro Santilli"

import os
import re
import unittest

from qgis.core import QgsDataSourceUri
from qgis.PyQt.QtCore import QObject
from qgis.testing import QgisTestCase, start_app
from qgis.utils import iface

start_app()

from db_manager.db_plugins.plugin import Table
from db_manager.db_plugins.postgis.connector import PostGisDBConnector
from db_manager.db_plugins.postgis.data_model import PGSqlResultModel
from db_manager.db_plugins.postgis.plugin import (
    PGDatabase,
    PGRasterTable,
    PostGisDBPlugin,
)


class TestDBManagerPostgisPlugin(QgisTestCase):
    @classmethod
    def setUpClass(self):
        self.old_pgdatabase_env = os.environ.get("PGDATABASE")
        # QGIS_PGTEST_DB contains the full connection string and not only the DB name!
        QGIS_PGTEST_DB = os.environ.get("QGIS_PGTEST_DB")
        if QGIS_PGTEST_DB is not None:
            test_uri = QgsDataSourceUri(QGIS_PGTEST_DB)
            self.testdb = test_uri.database()
        else:
            self.testdb = "qgis_test"
        os.environ["PGDATABASE"] = self.testdb

        # Create temporary service file
        self.old_pgservicefile_env = os.environ.get("PGSERVICEFILE")
        self.tmpservicefile = f"/tmp/qgis-test-{os.getpid()}-pg_service.conf"
        os.environ["PGSERVICEFILE"] = self.tmpservicefile

        f = open(self.tmpservicefile, "w")
        f.write(f"[dbmanager]\ndbname={self.testdb}\n")
        # TODO: add more things if PGSERVICEFILE was already set ?
        f.close()

    @classmethod
    def tearDownClass(self):
        # Restore previous env variables if needed
        if self.old_pgdatabase_env:
            os.environ["PGDATABASE"] = self.old_pgdatabase_env
        if self.old_pgservicefile_env:
            os.environ["PGSERVICEFILE"] = self.old_pgservicefile_env
        # Remove temporary service file
        os.unlink(self.tmpservicefile)

    # See https://github.com/qgis/QGIS/issues/24525

    def test_rasterTableURI(self):

        def check_rasterTableURI(expected_dbname):
            tables = database.tables()
            raster_tables_count = 0
            for tab in tables:
                if tab.type == Table.RasterType:
                    raster_tables_count += 1
                    uri = tab.uri()
                    m = re.search(" dbname='([^ ]*)' ", uri)
                    self.assertTrue(m)
                    actual_dbname = m.group(1)
                    self.assertEqual(actual_dbname, expected_dbname)
                # print(tab.type)
                # print(tab.quotedName())
                # print(tab)

            # We need to make sure a database is created with at
            # least one raster table !
            self.assertGreaterEqual(raster_tables_count, 1)

        obj = QObject()  # needs to be kept alive
        obj.connectionName = lambda: "fake"
        obj.providerName = lambda: "postgres"

        # Test for empty URI
        # See https://github.com/qgis/QGIS/issues/24525
        # and https://github.com/qgis/QGIS/issues/19005

        expected_dbname = self.testdb
        os.environ["PGDATABASE"] = expected_dbname

        database = PGDatabase(obj, QgsDataSourceUri())
        self.assertIsInstance(database, PGDatabase)

        uri = database.uri()
        self.assertEqual(uri.host(), "")
        self.assertEqual(uri.username(), "")
        self.assertEqual(uri.database(), expected_dbname)
        self.assertEqual(uri.service(), "")

        check_rasterTableURI(expected_dbname)

        # Test for service-only URI
        # See https://github.com/qgis/QGIS/issues/24526

        os.environ["PGDATABASE"] = "fake"
        database = PGDatabase(obj, QgsDataSourceUri("service=dbmanager"))
        self.assertIsInstance(database, PGDatabase)

        uri = database.uri()
        self.assertEqual(uri.host(), "")
        self.assertEqual(uri.username(), "")
        self.assertEqual(uri.database(), "")
        self.assertEqual(uri.service(), "dbmanager")

        check_rasterTableURI(expected_dbname)

    # See https://github.com/qgis/QGIS/issues/24732
    def test_unicodeInQuery(self):
        os.environ["PGDATABASE"] = self.testdb
        obj = QObject()  # needs to be kept alive
        obj.connectionName = lambda: "fake"
        obj.providerName = lambda: "postgres"
        database = PGDatabase(obj, QgsDataSourceUri())
        self.assertIsInstance(database, PGDatabase)
        # SQL as string literal
        res = database.sqlResultModel("SELECT 'é'::text", obj)
        self.assertIsInstance(res, PGSqlResultModel)
        dat = res.getData(0, 0)
        self.assertEqual(dat, "é")
        # SQL as unicode literal
        res = database.sqlResultModel("SELECT 'é'::text", obj)
        self.assertIsInstance(res, PGSqlResultModel)
        dat = res.getData(0, 0)
        self.assertEqual(dat, "é")


if __name__ == "__main__":
    unittest.main()



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.PyQt.QtWidgets import QApplication

from ..html_elems import HtmlTable
from ..info_model import DatabaseInfo


class GPKGDatabaseInfo(DatabaseInfo):
    def __init__(self, db):
        self.db = db

    def connectionDetails(self):
        tbl = [
            (
                QApplication.translate("DBManagerPlugin", "Filename:"),
                self.db.connector.dbname,
            )
        ]
        return HtmlTable(tbl)

    def generalInfo(self):
        return None

    def spatialInfo(self):
        return None

    def privilegesDetails(self):
        return None



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

# this will disable the dbplugin if the connector raise an ImportError
import re

from qgis.core import Qgis, QgsApplication, QgsSettings
from qgis.gui import QgsMessageBar
from qgis.PyQt.QtCore import QCoreApplication, QRegularExpression, Qt
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QApplication, QMessageBox

from ..plugin import (
    ConnectionError,
    Database,
    DBPlugin,
    InvalidDataException,
    RasterTable,
    Schema,
    Table,
    TableConstraint,
    TableField,
    TableIndex,
    TableRule,
    TableTrigger,
    VectorTable,
)
from .connector import PostGisDBConnector


def classFactory():
    return PostGisDBPlugin


class PostGisDBPlugin(DBPlugin):
    @classmethod
    def icon(self):
        return QgsApplication.getThemeIcon("/mIconPostgis.svg")

    @classmethod
    def typeName(self):
        return "postgis"

    @classmethod
    def typeNameString(self):
        return QCoreApplication.translate("db_manager", "PostgreSQL")

    @classmethod
    def providerName(self):
        return "postgres"

    @classmethod
    def connectionSettingsKey(self):
        return "/PostgreSQL/connections"

    def databasesFactory(self, connection, uri):
        return PGDatabase(connection, uri)

    def connect(self, parent=None):
        conn_name = self.connectionName()
        settings = QgsSettings()
        settings.beginGroup(f"/{self.connectionSettingsKey()}/{conn_name}")

        if not settings.contains("database"):  # non-existent entry?
            raise InvalidDataException(
                self.tr('There is no defined database connection "{0}".').format(
                    conn_name
                )
            )

        from qgis.core import QgsDataSourceUri

        uri = QgsDataSourceUri()

        settingsList = [
            "service",
            "host",
            "port",
            "database",
            "username",
            "password",
            "authcfg",
            "session_role",
        ]
        service, host, port, database, username, password, authcfg, session_role = (
            settings.value(x, "", type=str) for x in settingsList
        )

        useEstimatedMetadata = settings.value("estimatedMetadata", False, type=bool)
        try:
            sslmode = settings.enumValue("sslmode", QgsDataSourceUri.SslMode.SslPrefer)
        except TypeError:
            sslmode = QgsDataSourceUri.SslMode.SslPrefer

        settings.endGroup()

        if hasattr(authcfg, "isNull") and authcfg.isNull():
            authcfg = ""

        if service:
            uri.setConnection(service, database, username, password, sslmode, authcfg)
        else:
            uri.setConnection(
                host, port, database, username, password, sslmode, authcfg
            )

        uri.setUseEstimatedMetadata(useEstimatedMetadata)
        if session_role:
            uri.setParam("session_role", session_role)

        try:
            return self.connectToUri(uri)
        except ConnectionError:
            return False


class PGDatabase(Database):
    def __init__(self, connection, uri):
        Database.__init__(self, connection, uri)

    def connectorsFactory(self, uri):
        return PostGisDBConnector(uri, self.connection())

    def dataTablesFactory(self, row, db, schema=None):
        return PGTable(row, db, schema)

    def info(self):
        from .info_model import PGDatabaseInfo

        return PGDatabaseInfo(self)

    def vectorTablesFactory(self, row, db, schema=None):
        return PGVectorTable(row, db, schema)

    def rasterTablesFactory(self, row, db, schema=None):
        return PGRasterTable(row, db, schema)

    def schemasFactory(self, row, db):
        return PGSchema(row, db)

    def sqlResultModel(self, sql, parent):
        from .data_model import PGSqlResultModel

        return PGSqlResultModel(self, sql, parent)

    def sqlResultModelAsync(self, sql, parent):
        from .data_model import PGSqlResultModelAsync

        return PGSqlResultModelAsync(self, sql, parent)

    def registerDatabaseActions(self, mainWindow):
        Database.registerDatabaseActions(self, mainWindow)

        # add a separator
        separator = QAction(self)
        separator.setSeparator(True)
        mainWindow.registerAction(separator, self.tr("&Table"))

        action = QAction(self.tr("Run &Vacuum Analyze"), self)
        mainWindow.registerAction(
            action, self.tr("&Table"), self.runVacuumAnalyzeActionSlot
        )

        action = QAction(self.tr("Run &Refresh Materialized View"), self)
        mainWindow.registerAction(
            action, self.tr("&Table"), self.runRefreshMaterializedViewSlot
        )

    def runVacuumAnalyzeActionSlot(self, item, action, parent):
        QApplication.restoreOverrideCursor()
        try:
            if not isinstance(item, Table) or item.isView:
                parent.infoBar.pushMessage(
                    self.tr("Select a table for vacuum analyze."),
                    Qgis.MessageLevel.Info,
                    parent.iface.messageTimeout(),
                )
                return
        finally:
            QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)

        item.runVacuumAnalyze()

    def runRefreshMaterializedViewSlot(self, item, action, parent):
        QApplication.restoreOverrideCursor()
        try:
            if not isinstance(item, PGTable) or item._relationType != "m":
                parent.infoBar.pushMessage(
                    self.tr("Select a materialized view for refresh."),
                    Qgis.MessageLevel.Info,
                    parent.iface.messageTimeout(),
                )
                return
        finally:
            QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)

        item.runRefreshMaterializedView()

    def hasLowercaseFieldNamesOption(self):
        return True

    def supportsComment(self):
        return True

    def executeSql(self, sql):
        return self.connector._executeSql(sql)


class PGSchema(Schema):
    def __init__(self, row, db):
        Schema.__init__(self, db)
        self.oid, self.name, self.owner, self.perms, self.comment = row


class PGTable(Table):
    def __init__(self, row, db, schema=None):
        Table.__init__(self, db, schema)
        (
            self.name,
            schema_name,
            self._relationType,
            self.owner,
            self.estimatedRowCount,
            self.pages,
            self.comment,
        ) = row
        self.isView = self._relationType in {"v", "m"}
        self.estimatedRowCount = int(self.estimatedRowCount)

    def runVacuumAnalyze(self):
        self.aboutToChange.emit()
        self.database().connector.runVacuumAnalyze((self.schemaName(), self.name))
        # TODO: change only this item, not re-create all the tables in the schema/database
        self.schema().refresh() if self.schema() else self.database().refresh()

    def runRefreshMaterializedView(self):
        self.aboutToChange.emit()
        self.database().connector.runRefreshMaterializedView(
            (self.schemaName(), self.name)
        )
        # TODO: change only this item, not re-create all the tables in the schema/database
        self.schema().refresh() if self.schema() else self.database().refresh()

    def runAction(self, action):
        action = str(action)

        if action.startswith("vacuumanalyze/"):
            if action == "vacuumanalyze/run":
                self.runVacuumAnalyze()
                return True

        elif action.startswith("rule/"):
            parts = action.split("/")
            rule_name = parts[1]
            rule_action = parts[2]

            msg = self.tr("Do you want to {0} rule {1}?").format(rule_action, rule_name)

            QApplication.restoreOverrideCursor()

            try:
                if (
                    QMessageBox.question(
                        None,
                        self.tr("Table rule"),
                        msg,
                        QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    )
                    == QMessageBox.StandardButton.No
                ):
                    return False
            finally:
                QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)

            if rule_action == "delete":
                self.aboutToChange.emit()
                self.database().connector.deleteTableRule(
                    rule_name, (self.schemaName(), self.name)
                )
                self.refreshRules()
                return True

        elif action.startswith("refreshmaterializedview/"):
            if action == "refreshmaterializedview/run":
                self.runRefreshMaterializedView()
                return True

        return Table.runAction(self, action)

    def tableFieldsFactory(self, row, table):
        return PGTableField(row, table)

    def tableConstraintsFactory(self, row, table):
        return PGTableConstraint(row, table)

    def tableIndexesFactory(self, row, table):
        return PGTableIndex(row, table)

    def tableTriggersFactory(self, row, table):
        return PGTableTrigger(row, table)

    def tableRulesFactory(self, row, table):
        return PGTableRule(row, table)

    def info(self):
        from .info_model import PGTableInfo

        return PGTableInfo(self)

    def crs(self):
        return self.database().connector.getCrs(self.srid)

    def tableDataModel(self, parent):
        from .data_model import PGTableDataModel

        return PGTableDataModel(self, parent)

    def delete(self):
        self.aboutToChange.emit()
        if self.isView:
            ret = self.database().connector.deleteView(
                (self.schemaName(), self.name), self._relationType == "m"
            )
        else:
            ret = self.database().connector.deleteTable((self.schemaName(), self.name))
        if not ret:
            self.deleted.emit()
        return ret


class PGVectorTable(PGTable, VectorTable):
    def __init__(self, row, db, schema=None):
        PGTable.__init__(self, row[:-4], db, schema)
        VectorTable.__init__(self, db, schema)
        self.geomColumn, self.geomType, self.geomDim, self.srid = row[-4:]

    def info(self):
        from .info_model import PGVectorTableInfo

        return PGVectorTableInfo(self)

    def runAction(self, action):
        if PGTable.runAction(self, action):
            return True
        return VectorTable.runAction(self, action)

    def geometryType(self):
        """Returns the proper WKT type.
        PostGIS records type like this:
        | WKT Type     | geomType    | geomDim |
        |--------------|-------------|---------|
        | LineString   | LineString  | 2       |
        | LineStringZ  | LineString  | 3       |
        | LineStringM  | LineStringM | 3       |
        | LineStringZM | LineString  | 4       |
        """
        geometryType = self.geomType
        if self.geomDim == 3 and self.geomType[-1] != "M":
            geometryType += "Z"
        elif self.geomDim == 4:
            geometryType += "ZM"

        return geometryType


class PGRasterTable(PGTable, RasterTable):
    def __init__(self, row, db, schema=None):
        PGTable.__init__(self, row[:-6], db, schema)
        RasterTable.__init__(self, db, schema)
        (
            self.geomColumn,
            self.pixelType,
            self.pixelSizeX,
            self.pixelSizeY,
            self.isExternal,
            self.srid,
        ) = row[-6:]
        self.geomType = "RASTER"

    def info(self):
        from .info_model import PGRasterTableInfo

        return PGRasterTableInfo(self)

    def uri(self, uri=None):
        """Returns the datasource URI for postgresraster provider"""

        if not uri:
            uri = self.database().uri()
        service = ("service='%s'" % uri.service()) if uri.service() else ""
        dbname = ("dbname='%s'" % uri.database()) if uri.database() else ""
        host = ("host=%s" % uri.host()) if uri.host() else ""
        user = ("user=%s" % uri.username()) if uri.username() else ""
        passw = ("password=%s" % uri.password()) if uri.password() else ""
        port = ("port=%s" % uri.port()) if uri.port() else ""

        schema = self.schemaName() if self.schemaName() else "public"
        table = f'"{schema}"."{self.name}"'

        if not dbname:
            # postgresraster provider *requires* a dbname
            connector = self.database().connector
            r = connector._execute(None, "SELECT current_database()")
            dbname = "dbname='%s'" % connector._fetchone(r)[0]
            connector._close_cursor(r)

        # Find first raster field
        col = ""
        for fld in self.fields():
            if fld.dataType == "raster":
                col = "column='%s'" % fld.name
                break

        uri = f"{service} {dbname} {host} {user} {passw} {port} {col} table={table}"

        return uri

    def mimeUri(self):
        uri = "raster:postgresraster:{}:{}".format(
            self.name, re.sub(":", r"\:", self.uri())
        )
        return uri

    def toMapLayer(self, geometryType=None, crs=None):
        from qgis.core import (
            QgsContrastEnhancement,
            QgsCredentials,
            QgsDataSourceUri,
            QgsRasterLayer,
        )

        rl = QgsRasterLayer(self.uri(), self.name, "postgresraster")
        if not rl.isValid():
            err = rl.error().summary()
            uri = QgsDataSourceUri(self.database().uri())
            conninfo = uri.connectionInfo(False)
            username = uri.username()
            password = uri.password()

            for i in range(3):
                (ok, username, password) = QgsCredentials.instance().get(
                    conninfo, username, password, err
                )
                if ok:
                    uri.setUsername(username)
                    uri.setPassword(password)
                    rl = QgsRasterLayer(self.uri(uri), self.name)
                    if rl.isValid():
                        break

        if rl.isValid():
            rl.setContrastEnhancement(
                QgsContrastEnhancement.ContrastEnhancementAlgorithm.StretchToMinimumMaximum
            )
        return rl


class PGTableField(TableField):
    def __init__(self, row, table):
        TableField.__init__(self, table)
        (
            self.num,
            self.name,
            self.dataType,
            self.charMaxLen,
            self.modifier,
            self.notNull,
            self.hasDefault,
            self.default,
            typeStr,
        ) = row
        self.primaryKey = False

        # get modifier (e.g. "precision,scale") from formatted type string
        trimmedTypeStr = typeStr.strip()
        regex = QRegularExpression(r"\((.+)\)$")
        match = regex.match(trimmedTypeStr)
        if match.hasMatch():
            self.modifier = match.captured(1).strip()
        else:
            self.modifier = None

        # find out whether fields are part of primary key
        for con in self.table().constraints():
            if con.type == TableConstraint.TypePrimaryKey and self.num in con.columns:
                self.primaryKey = True
                break

    def getComment(self):
        """Returns the comment for a field"""
        tab = self.table()
        # SQL Query checking if a comment exists for the field
        sql_cpt = f"Select count(*) from pg_description pd, pg_class pc, pg_attribute pa where relname = '{tab.name}' and attname = '{self.name}' and pa.attrelid = pc.oid and pd.objoid = pc.oid and pd.objsubid = pa.attnum"
        # SQL Query that return the comment of the field
        sql = f"Select pd.description from pg_description pd, pg_class pc, pg_attribute pa where relname = '{tab.name}' and attname = '{self.name}' and pa.attrelid = pc.oid and pd.objoid = pc.oid and pd.objsubid = pa.attnum"
        c = tab.database().connector._execute(None, sql_cpt)  # Execute Check query
        res = tab.database().connector._fetchone(c)[0]  # Store result
        if res == 1:
            # When a comment exists
            c = tab.database().connector._execute(None, sql)  # Execute query
            res = tab.database().connector._fetchone(c)[0]  # Store result
            tab.database().connector._close_cursor(c)  # Close cursor
            return res  # Return comment
        else:
            return ""


class PGTableConstraint(TableConstraint):
    def __init__(self, row, table):
        TableConstraint.__init__(self, table)
        self.name, constr_type_str, self.isDefferable, self.isDeffered, columns = row[
            :5
        ]
        self.columns = list(map(int, columns.split(" ")))

        if constr_type_str in TableConstraint.types:
            self.type = TableConstraint.types[constr_type_str]
        else:
            self.type = TableConstraint.TypeUnknown

        if self.type == TableConstraint.TypeCheck:
            self.checkSource = row[5]
        elif self.type == TableConstraint.TypeForeignKey:
            self.foreignTable = row[6]
            self.foreignOnUpdate = TableConstraint.onAction[row[7]]
            self.foreignOnDelete = TableConstraint.onAction[row[8]]
            self.foreignMatchType = TableConstraint.matchTypes[row[9]]
            self.foreignKeys = row[10]


class PGTableIndex(TableIndex):
    def __init__(self, row, table):
        TableIndex.__init__(self, table)
        self.name, columns, self.isUnique = row
        self.columns = list(map(int, columns.split(" ")))


class PGTableTrigger(TableTrigger):
    def __init__(self, row, table):
        TableTrigger.__init__(self, table)
        self.name, self.function, self.type, self.enabled = row


class PGTableRule(TableRule):
    def __init__(self, row, table):
        TableRule.__init__(self, table)
        self.name, self.definition = row



"""
***************************************************************************
    i_maxlik.py
    -----------
    Date                 : March 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "March 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

from .i import importSigFile, regroupRasters


def processCommand(alg, parameters, context, feedback):
    group, subgroup = regroupRasters(
        alg, parameters, context, "input", "group", "subgroup"
    )

    # import signature
    signatureFile = alg.parameterAsString(parameters, "signaturefile", context)
    shortSigFile = importSigFile(alg, group, subgroup, signatureFile)
    parameters["signaturefile"] = shortSigFile

    # Handle other parameters
    alg.processCommand(parameters, context, feedback)



# -*- coding: utf-8 -*-

"""
***************************************************************************
    QtNetwork.py
    ---------------------
    Date                 : March 2016
    Copyright            : (C) 2016 by Juergen E. Fischer
    Email                : jef at norbit dot de
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = 'Juergen E. Fischer'
__date__ = 'March 2016'
__copyright__ = '(C) 2016, Juergen E. Fischer'

from PyQt6.QtNetwork import *
from .common import install_extended_enum_wrapper

if 6 == 6:
    # patch back in Qt flags removed in PyQt
    QAbstractSocket.BindMode = lambda flags=0: QAbstractSocket.BindFlag(flags)
    QAbstractSocket.PauseModes = lambda flags=0: QAbstractSocket.PauseMode(flags)
    QHostAddress.ConversionMode = lambda flags=0: QHostAddress.ConversionModeFlag(flags)
    QHstsPolicy.PolicyFlags = lambda flags=0: QHstsPolicy.PolicyFlag(flags)
    QLocalServer.SocketOptions = lambda flags=0: QLocalServer.SocketOption(flags)
    QNetworkInterface.InterfaceFlags = lambda flags=0: QNetworkInterface.InterfaceFlag(flags)
    QNetworkProxy.Capabilities = lambda flags=0: QNetworkProxy.Capability(flags)
    QSsl.SslOptions = lambda flags=0: QSsl.SslOption(flags)

    QNetworkRequest.setAttribute = install_extended_enum_wrapper(QNetworkRequest.setAttribute)




"""
/***************************************************************************
Python Console for QGIS
                             -------------------
begin                : 2012-09-10
copyright            : (C) 2012 by Salvatore Larosa
email                : lrssvtml (at) gmail (dot) com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
Some portions of code were taken from https://code.google.com/p/pydee/
"""

from pathlib import Path

from qgis.core import Qgis, QgsApplication, QgsSettings, QgsSettingsTree
from qgis.gui import QgsOptionsPageWidget, QgsOptionsWidgetFactory
from qgis.PyQt import uic
from qgis.PyQt.QtCore import QCoreApplication, QUrl
from qgis.PyQt.QtGui import QDesktopServices, QIcon
from qgis.PyQt.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QMessageBox,
    QTableWidgetItem,
    QWidget,
)

from .console_compile_apis import PrepareAPIDialog

Ui_SettingsDialogPythonConsole, _ = uic.loadUiType(
    Path(__file__).parent / "console_settings.ui"
)


class ConsoleOptionsFactory(QgsOptionsWidgetFactory):
    def __init__(self):
        super(QgsOptionsWidgetFactory, self).__init__()

    def icon(self):
        return QgsApplication.getThemeIcon("/console/mIconRunConsole.svg")

    def path(self):
        return ["ide"]

    def createWidget(self, parent):
        return ConsoleOptionsPage(parent)


class ConsoleOptionsPage(QgsOptionsPageWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.options_widget = ConsoleOptionsWidget(parent)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setMargin(0)
        self.setLayout(layout)
        layout.addWidget(self.options_widget)
        self.setObjectName("consoleOptions")

    def apply(self):
        self.options_widget.accept()

    def helpKey(self):
        return "plugins/python_console.html"


class ConsoleOptionsWidget(QWidget, Ui_SettingsDialogPythonConsole):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle(
            QCoreApplication.translate(
                "SettingsDialogPythonConsole", "Python Console Settings"
            )
        )
        self.parent = parent
        self.setupUi(self)

        # Populate the documentation Browser combobox
        self.contextHelpBrowser.addItem(
            QCoreApplication.translate(
                "PythonConsole", "Embedded Webview (developer tools)"
            ),
            Qgis.DocumentationBrowser.DeveloperToolsPanel,
        )
        self.contextHelpBrowser.addItem(
            QCoreApplication.translate("PythonConsole", "Default System Web Browser"),
            Qgis.DocumentationBrowser.SystemWebBrowser,
        )

        self.autopep8Level.setClearValue(1)
        self.maxLineLength.setClearValue(80)

        # Set up the formatter combo box
        self.formatter.addItems(["autopep8", "black"])

        self.listPath = []
        self.lineEdit.setReadOnly(True)

        self.restoreSettings()
        self.initialCheck()

        self.addAPIpath.setIcon(QIcon(":/images/themes/default/symbologyAdd.svg"))
        self.addAPIpath.setToolTip(
            QCoreApplication.translate("PythonConsole", "Add API path")
        )
        self.removeAPIpath.setIcon(QIcon(":/images/themes/default/symbologyRemove.svg"))
        self.removeAPIpath.setToolTip(
            QCoreApplication.translate("PythonConsole", "Remove API path")
        )

        self.preloadAPI.stateChanged.connect(self.initialCheck)
        self.addAPIpath.clicked.connect(self.loadAPIFile)
        self.removeAPIpath.clicked.connect(self.removeAPI)
        self.compileAPIs.clicked.connect(self._prepareAPI)

        self.formatter.currentTextChanged.connect(self.onFormatterChanged)
        self.onFormatterChanged()

    def initialCheck(self):
        if self.preloadAPI.isChecked():
            self.enableDisable(False)
        else:
            self.enableDisable(True)

    def enableDisable(self, value):
        self.tableWidget.setEnabled(value)
        self.addAPIpath.setEnabled(value)
        self.removeAPIpath.setEnabled(value)
        self.groupBoxPreparedAPI.setEnabled(value)

    def loadAPIFile(self):
        settings = QgsSettings()
        lastDirPath = settings.value("pythonConsole/lastDirAPIPath", "", type=str)
        fileAPI, selected_filter = QFileDialog.getOpenFileName(
            self, "Open API File", lastDirPath, "API file (*.api)"
        )
        if fileAPI:
            self.addAPI(fileAPI)
            settings.setValue("pythonConsole/lastDirAPIPath", fileAPI)

    def _prepareAPI(self):
        if self.tableWidget.rowCount() != 0:
            pap_file, filter = QFileDialog().getSaveFileName(
                self, "", "*.pap", "Prepared APIs file (*.pap)"
            )
        else:
            QMessageBox.information(
                self,
                self.tr("Warning!"),
                self.tr("You need to add some APIs file in order to compile"),
            )
            return
        if pap_file:
            api_lexer = "QsciLexerPython"
            api_files = []
            count = self.tableWidget.rowCount()
            for i in range(0, count):
                api_files.append(self.tableWidget.item(i, 1).text())
            api_dlg = PrepareAPIDialog(api_lexer, api_files, pap_file, self)
            api_dlg.show()
            api_dlg.activateWindow()
            api_dlg.raise_()
            api_dlg.prepareAPI()
            self.lineEdit.setText(pap_file)

    def accept(self):
        if not self.preloadAPI.isChecked() and not self.groupBoxPreparedAPI.isChecked():
            if self.tableWidget.rowCount() == 0:
                QMessageBox.information(
                    self,
                    self.tr("Warning!"),
                    self.tr(
                        'Please specify API file or check "Use preloaded API files"'
                    ),
                )
                return
        if self.groupBoxPreparedAPI.isChecked() and not self.lineEdit.text():
            QMessageBox.information(
                self,
                self.tr("Warning!"),
                QCoreApplication.translate(
                    "optionsDialog",
                    'The APIs file was not compiled, click on "Compile APIs…"',
                ),
            )
            return
        self.saveSettings()
        self.listPath = []

    def addAPI(self, pathAPI):
        count = self.tableWidget.rowCount()
        self.tableWidget.setColumnCount(2)
        self.tableWidget.insertRow(count)
        pathItem = QTableWidgetItem(pathAPI)
        pathSplit = pathAPI.split("/")
        apiName = pathSplit[-1][0:-4]
        apiNameItem = QTableWidgetItem(apiName)
        self.tableWidget.setItem(count, 0, apiNameItem)
        self.tableWidget.setItem(count, 1, pathItem)

    def removeAPI(self):
        listItemSel = self.tableWidget.selectionModel().selectedRows()
        for index in reversed(listItemSel):
            self.tableWidget.removeRow(index.row())

    def saveSettings(self):
        settings = QgsSettings()
        settings.setValue("pythonConsole/preloadAPI", self.preloadAPI.isChecked())
        settings.setValue(
            "pythonConsole/autoSaveScript", self.autoSaveScript.isChecked()
        )

        for i in range(0, self.tableWidget.rowCount()):
            text = self.tableWidget.item(i, 1).text()
            self.listPath.append(text)
        settings.setValue("pythonConsole/userAPI", self.listPath)

        settings.setValue(
            "pythonConsole/autoCompThreshold", self.autoCompThreshold.value()
        )
        settings.setValue(
            "pythonConsole/autoCompleteEnabled", self.groupBoxAutoCompletion.isChecked()
        )

        settings.setValue(
            "pythonConsole/usePreparedAPIFile", self.groupBoxPreparedAPI.isChecked()
        )
        settings.setValue("pythonConsole/preparedAPIFile", self.lineEdit.text())

        if self.autoCompFromAPI.isChecked():
            settings.setValue("pythonConsole/autoCompleteSource", "fromAPI")
        elif self.autoCompFromDoc.isChecked():
            settings.setValue("pythonConsole/autoCompleteSource", "fromDoc")
        elif self.autoCompFromDocAPI.isChecked():
            settings.setValue("pythonConsole/autoCompleteSource", "fromDocAPI")

        settings.setValue(
            "pythonConsole/enableObjectInsp", self.enableObjectInspector.isChecked()
        )
        settings.setValue(
            "pythonConsole/autoCloseBracket", self.autoCloseBracket.isChecked()
        )
        settings.setValue("pythonConsole/autoSurround", self.autoSurround.isChecked())
        settings.setValue(
            "pythonConsole/autoInsertImport", self.autoInsertImport.isChecked()
        )

        settings.setValue("pythonConsole/formatOnSave", self.formatOnSave.isChecked())

        codeEditorTreeNode = QgsSettingsTree.node("gui").childNode("code-editor")
        pythonSettingsTreeNode = codeEditorTreeNode.childNode("python")
        pythonSettingsTreeNode.childSetting("sort-imports").setValue(
            self.sortImports.isChecked()
        )
        pythonSettingsTreeNode.childSetting("formatter").setValue(
            self.formatter.currentText()
        )
        pythonSettingsTreeNode.childSetting("autopep8-level").setValue(
            self.autopep8Level.value()
        )
        pythonSettingsTreeNode.childSetting("black-normalize-quotes").setValue(
            self.blackNormalizeQuotes.isChecked()
        )
        pythonSettingsTreeNode.childSetting("max-line-length").setValue(
            self.maxLineLength.value()
        )
        pythonSettingsTreeNode.childSetting("external-editor").setValue(
            self.externalEditor.text()
        )
        pythonSettingsTreeNode.childSetting("context-help-browser").setVariantValue(
            self.contextHelpBrowser.currentData().name
        )

        codeEditorTreeNode.childSetting("context-help-hover").setValue(
            self.contextHelpHover.isChecked()
        )

    def restoreSettings(self):
        settings = QgsSettings()
        self.preloadAPI.setChecked(
            settings.value("pythonConsole/preloadAPI", True, type=bool)
        )
        self.lineEdit.setText(
            settings.value("pythonConsole/preparedAPIFile", "", type=str)
        )
        itemTable = settings.value("pythonConsole/userAPI", [])
        if itemTable:
            self.tableWidget.setRowCount(0)
            for i in range(len(itemTable)):
                self.tableWidget.insertRow(i)
                self.tableWidget.setColumnCount(2)
                pathSplit = itemTable[i].split("/")
                apiName = pathSplit[-1][0:-4]
                self.tableWidget.setItem(i, 0, QTableWidgetItem(apiName))
                self.tableWidget.setItem(i, 1, QTableWidgetItem(itemTable[i]))
        self.autoSaveScript.setChecked(
            settings.value("pythonConsole/autoSaveScript", False, type=bool)
        )

        self.autoCompThreshold.setValue(
            settings.value("pythonConsole/autoCompThreshold", 2, type=int)
        )
        self.groupBoxAutoCompletion.setChecked(
            settings.value("pythonConsole/autoCompleteEnabled", True, type=bool)
        )

        self.enableObjectInspector.setChecked(
            settings.value("pythonConsole/enableObjectInsp", False, type=bool)
        )
        self.autoCloseBracket.setChecked(
            settings.value("pythonConsole/autoCloseBracket", True, type=bool)
        )
        self.autoSurround.setChecked(
            settings.value("pythonConsole/autoSurround", True, type=bool)
        )
        self.autoInsertImport.setChecked(
            settings.value("pythonConsole/autoInsertImport", False, type=bool)
        )

        codeEditorTreeNode = QgsSettingsTree.node("gui").childNode("code-editor")
        pythonSettingsTreeNode = codeEditorTreeNode.childNode("python")

        self.formatOnSave.setChecked(
            settings.value("pythonConsole/formatOnSave", False, type=bool)
        )
        self.sortImports.setChecked(
            pythonSettingsTreeNode.childSetting("sort-imports").value()
        )
        self.formatter.setCurrentText(
            pythonSettingsTreeNode.childSetting("formatter").value()
        )
        self.autopep8Level.setValue(
            pythonSettingsTreeNode.childSetting("autopep8-level").value()
        )
        self.blackNormalizeQuotes.setChecked(
            pythonSettingsTreeNode.childSetting("black-normalize-quotes").value()
        )
        self.maxLineLength.setValue(
            pythonSettingsTreeNode.childSetting("max-line-length").value()
        )

        browserName = pythonSettingsTreeNode.childSetting(
            "context-help-browser"
        ).valueAsVariant()
        try:
            browser = Qgis.DocumentationBrowser[browserName]
        except KeyError:
            browser = Qgis.DocumentationBrowser.DeveloperToolsPanel

        self.contextHelpBrowser.setCurrentIndex(
            self.contextHelpBrowser.findData(browser)
        )
        self.contextHelpHover.setChecked(
            codeEditorTreeNode.childSetting("context-help-hover").value()
        )

        if settings.value("pythonConsole/autoCompleteSource") == "fromDoc":
            self.autoCompFromDoc.setChecked(True)
        elif settings.value("pythonConsole/autoCompleteSource") == "fromAPI":
            self.autoCompFromAPI.setChecked(True)
        elif settings.value("pythonConsole/autoCompleteSource") == "fromDocAPI":
            self.autoCompFromDocAPI.setChecked(True)

        self.externalEditor.setText(
            pythonSettingsTreeNode.childSetting("external-editor").value()
        )

    def onFormatterChanged(self):
        """Toggle formatter-specific options visibility when the formatter is changed"""
        if self.formatter.currentText() == "autopep8":
            self.autopep8Level.setVisible(True)
            self.autopep8LevelLabel.setVisible(True)
            self.blackNormalizeQuotes.setVisible(False)
        else:  # black
            self.autopep8Level.setVisible(False)
            self.autopep8LevelLabel.setVisible(False)
            self.blackNormalizeQuotes.setVisible(True)
